# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2022 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import numbers
from functools import reduce
from operator import add

import numpy as np

from .. import pointwise
from ..domain_tuple import DomainTuple
from ..logger import logger
from ..multi_domain import MultiDomain
from ..utilities import NiftyMeta, check_object_identity, indent, myassert


class Operator(metaclass=NiftyMeta):
    """Transforms values defined on one domain into values defined on another
    domain, and can also provide the Jacobian.
    """

    @property
    def domain(self):
        """The domain on which the Operator's input Field is defined.

        Returns
        -------
        domain : DomainTuple or MultiDomain
        """
        return self._domain

    @property
    def target(self):
        """The domain on which the Operator's output Field is defined.

        Returns
        -------
        target : DomainTuple or MultiDomain
        """
        return self._target

    @property
    def val(self):
        """The numerical value associated with this object
        For "pure" operators this is `None`. For Field-like objects this
        is a `numpy.ndarray` or a dictionary of `numpy.ndarray`s mathcing the
        object's `target`.

        Returns
        -------
        None or numpy.ndarray or dictionary of np.ndarrays : the numerical value
        """
        return None

    @property
    def jac(self):
        """The Jacobian associated with this object.
        For "pure" operators this is `None`. For Field-like objects this
        can be `None` (in which case the object is a constant), or it can be a
        `LinearOperator` with `domain` and `target` matching the object's.

        Returns
        -------
        None or LinearOperator : the Jacobian

        Notes
        -----
        if `value` is None, this must be `None` as well!
        """
        return None

    @property
    def want_metric(self):
        """Whether a metric should be computed for the full expression.
        This is `False` whenever `jac` is `None`. In other cases it signals
        that operators processing this object should compute the metric.

        Returns
        -------
        bool : whether the metric should be computed
        """
        return False

    @property
    def metric(self):
        """The metric associated with the object.
        This is `None`, except when all the following conditions hold:
        - `want_metric` is `True`

        - `target` is the scalar domain

        - the operator chain contained an operator which could compute the
          metric

        Returns
        -------
        None or LinearOperator : the metric
        """
        return None

    def isIdentity(self):  # Will be overloaded in ScalingOperator
        return False

    def scale(self, factor):
        if not isinstance(factor, numbers.Number):
            raise TypeError(".scale() takes a number as input")
        if factor == 1:
            return self
        from .scaling_operator import ScalingOperator
        return ScalingOperator(self.target, factor)(self)

    def conjugate(self):
        from .simple_linear_operators import ConjugationOperator
        return ConjugationOperator(self.target)(self)

    def sum(self, spaces=None):
        from .contraction_operator import ContractionOperator
        return ContractionOperator(self.target, spaces)(self)

    def integrate(self, spaces=None):
        from .contraction_operator import IntegrationOperator
        return IntegrationOperator(self.target, spaces)(self)

    def vdot(self, other):
        from ..sugar import makeOp
        if not isinstance(other, Operator):
            raise TypeError
        if other.jac is None:
            res = self.conjugate()*other
        else:
            res = makeOp(other) @ self.conjugate()
        return res.sum()

    @property
    def real(self):
        from .simple_linear_operators import Realizer
        return Realizer(self.target)(self)

    @property
    def imag(self):
        from .simple_linear_operators import Imaginizer
        return Imaginizer(self.target)(self)

    def __neg__(self):
        return self.scale(-1)

    def __matmul__(self, x):
        from .energy_operators import LikelihoodEnergyOperator

        if not isinstance(x, Operator):
            return NotImplemented
        if isinstance(x, LikelihoodEnergyOperator):
            return NotImplemented
        if x.target is self.domain:
            if x.isIdentity():
                return self
            if self.isIdentity():
                return x
            return _OpChain.make((self, x))
        return self.partial_insert(x)

    def __rmatmul__(self, x):
        from .energy_operators import LikelihoodEnergyOperator

        if not isinstance(x, Operator):
            return NotImplemented
        if isinstance(x, LikelihoodEnergyOperator):
            return NotImplemented
        if x.domain is self.target:
            if x.isIdentity():
                return self
            if self.isIdentity():
                return x
            return _OpChain.make((x, self))
        return x.partial_insert(self)

    def partial_insert(self, x):
        from ..multi_domain import MultiDomain
        if not isinstance(x, Operator):
            raise TypeError
        if not isinstance(self.domain, MultiDomain):
            raise TypeError
        if not isinstance(x.target, MultiDomain):
            raise TypeError
        bigdom = MultiDomain.union([self.domain, x.target])
        k1, k2 = set(self.domain.keys()), set(x.target.keys())
        le, ri = k2 - k1, k1 - k2
        leop, riop = self, x
        if len(ri) > 0:
            riop = riop + self.identity_operator(
                MultiDomain.make({kk: bigdom[kk]
                                  for kk in ri}))
        if len(le) > 0:
            leop = leop + self.identity_operator(
                MultiDomain.make({kk: bigdom[kk]
                                  for kk in le}))
        return leop @ riop

    @staticmethod
    def identity_operator(dom):
        from ..sugar import makeDomain
        from .block_diagonal_operator import BlockDiagonalOperator
        from .scaling_operator import ScalingOperator

        dom = makeDomain(dom)
        if isinstance(dom, DomainTuple):
            return ScalingOperator(dom, 1.)
        idops = {kk: ScalingOperator(dd, 1.) for kk, dd in dom.items()}
        return BlockDiagonalOperator(dom, idops)

    def __mul__(self, x):
        if isinstance(x, Operator):
            return _OpProd(self, x)
        if np.isscalar(x):
            return self.scale(x)
        return NotImplemented

    def __rmul__(self, x):
        return self.__mul__(x)

    def __add__(self, x):
        if not isinstance(x, Operator):
            return NotImplemented
        return _OpSum(self, x)

    def __sub__(self, x):
        if not isinstance(x, Operator):
            return NotImplemented
        return _OpSum(self, -x)

    def __abs__(self):
        return self.ptw("abs")

    def __pow__(self, power):
        if not (np.isscalar(power) or power.jac is None):
            return NotImplemented
        return self.ptw("power", power)

    def __getitem__(self, key):
        from ..sugar import is_operator
        from .simple_linear_operators import ducktape

        if not is_operator(self):
            raise TypeError(f"'{type(self)}' object is not subscriptable")
        if not isinstance(self.target, MultiDomain):
            raise TypeError("Only Operators with a MultiDomain as target can be subscripted.")
        return ducktape(None, self, key) @ self

    def apply(self, x):
        """Applies the operator to a Field, MultiField or Linearization.

        Parameters
        ----------
        x : :class:`nifty.cl.field.Field`, :class:`nifty.cl.multi_field.MultiField`,
            or :class:`nifty.cl.linearization.Linearization`
            Input on which the operator shall act. Needs to be defined on
            :attr:`domain`. If `x`is a :class:`nifty.cl.linearization.Linearization`,
            `apply` returns a new :class:`nifty.cl.linearization.Linearization`
            contining the result of the operator application as well as its
            Jacobian, evaluated at `x`.
        """
        raise NotImplementedError

    def force(self, x):
        """Extract subset of domain of x according to `self.domain` and apply
        operator."""
        return self.apply(x.extract(self.domain))

    def _check_input(self, x):
        from .scaling_operator import ScalingOperator
        if not (isinstance(x, Operator) and x.val is not None):
            raise TypeError
        if x.jac is not None:
            if not isinstance(x.jac, ScalingOperator):
                raise ValueError
            if x.jac._factor != 1:
                raise ValueError
        check_object_identity(self._domain, x.domain)

    def __call__(self, x):
        if not isinstance(x, Operator):
            raise TypeError
        if x.jac is not None:
            return self.apply(x.trivial_jac()).prepend_jac(x.jac)
        elif x.val is not None:
            return self.apply(x)
        return self @ x

    def ducktape(self, name):
        from ..sugar import is_operator, makeDomain
        from .simple_linear_operators import DomainChangerAndReshaper, ducktape

        if not is_operator(self):
            raise RuntimeError("ducktape works only on operators")

        if isinstance(name, str):  # convert to MultiDomain
            return self @ ducktape(self, None, name)
        else:  # convert domain
            newdom = makeDomain(name)
            return self @ DomainChangerAndReshaper(newdom, self.domain)

    def ducktape_left(self, name):
        from ..sugar import is_fieldlike, is_operator
        from .simple_linear_operators import DomainChangerAndReshaper, ducktape

        if isinstance(name, str):  # convert to MultiDomain
            tgt = self.target if is_operator(self) else self.domain
            return ducktape(None, tgt, name)(self)
        else:  # convert domain
            newdom = DomainTuple.make(name)
            dom = self.domain if is_fieldlike(self) else self.target
            return DomainChangerAndReshaper(dom, newdom)(self)

    def transpose(self, indices):
        """Transposes a Field.

        Parameters
        ----------
        indices : tuple
            Must be a tuple or list which contains a permutation of
            [0,1,..,N-1] where N is the number of domains in the target of the
            Operator (or the Field).
        """
        from ..sugar import is_fieldlike
        from .transpose_operator import TransposeOperator

        dom = self.domain if is_fieldlike(self) else self.target
        return TransposeOperator(dom, indices)(self)

    def __repr__(self):
        return self.__class__.__name__

    def simplify_for_constant_input(self, c_inp):
        from ..multi_field import MultiField
        from ..sugar import makeDomain
        from .energy_operators import EnergyOperator, LikelihoodEnergyOperator
        from .simplify_for_const import (ConstantEnergyOperator,
                                         ConstantLikelihoodEnergyOperator,
                                         ConstantOperator)
        if c_inp is None or (isinstance(c_inp, MultiField) and len(c_inp.keys()) == 0):
            return None, self
        dom = c_inp.domain
        if isinstance(dom, MultiDomain) and len(dom) == 0:
            return None, self

        # Convention: If c_inp is MultiField, it needs to be defined on a
        # subdomain of self._domain
        if isinstance(self.domain, MultiDomain):
            myassert(isinstance(dom, MultiDomain))
            if not set(c_inp.keys()) <= set(self.domain.keys()):
                raise ValueError

        if dom is self.domain:
            if isinstance(self, DomainTuple):
                raise RuntimeError
            if isinstance(self, EnergyOperator):
                op = ConstantEnergyOperator(self(c_inp))
            elif isinstance(self, LikelihoodEnergyOperator):
                op = ConstantLikelihoodEnergyOperator(self(c_inp))
            else:
                op = ConstantOperator(self(c_inp))
            return None, op
        if not isinstance(dom, MultiDomain):
            raise RuntimeError
        c_out, op = self._simplify_for_constant_input_nontrivial(c_inp)
        vardom = makeDomain({kk: vv for kk, vv in self.domain.items()
                             if kk not in c_inp.keys()})
        myassert(op.domain is vardom)
        myassert(op.target is self.target)
        myassert(isinstance(op, Operator))
        if c_out is not None:
            myassert(isinstance(c_out, MultiField))
            myassert(len(set(c_out.keys()) & self.domain.keys()) == 0)
            myassert(set(c_out.keys()) <= set(c_inp.keys()))
        return c_out, op

    def _simplify_for_constant_input_nontrivial(self, c_inp):
        from .simplify_for_const import InsertionOperator
        logger.warning('SlowPartialConstantOperator used for:')
        logger.warning(self.__repr__())
        return None, self @ InsertionOperator(self.domain, c_inp)

    def ptw(self, op, *args, **kwargs):
        return _OpChain.make((_FunctionApplier(self.target, op, *args, **kwargs), self))

    def ptw_pre(self, op, *args, **kwargs):
        return _OpChain.make((self, _FunctionApplier(self.domain, op, *args, **kwargs)))

    def apply_to_random_sample(self, **kwargs):
        """Applies the operator to a sample drawn with `ift.sugar.from_random`.
        Keyword arguments are passed through to the sample generation.

        .. warning::
            If not specified otherwise, the sample will be of dtype
            `np.float64`. If the operator requires input values of other
            dtypes, this needs to be indicated with the `dtype` keyword argument.
        """
        from ..sugar import from_random
        random_input = from_random(self.domain, **kwargs)
        return self(random_input)


for f in pointwise.ptw_dict.keys():
    def func(f):
        def func2(self, *args, **kwargs):
            return self.ptw(f, *args, **kwargs)
        return func2
    setattr(Operator, f, func(f))
    def func(f):
        def func2(self, *args, **kwargs):
            return self.ptw_pre(f, *args, **kwargs)
        return func2
    setattr(Operator, f + "_pre", func(f))


class _FunctionApplier(Operator):
    def __init__(self, domain, funcname, *args, **kwargs):
        from ..sugar import makeDomain
        self._domain = self._target = makeDomain(domain)
        self._funcname = funcname
        self._args = args
        self._kwargs = kwargs

    def _device_preparation(self, x):
        from ..any_array import AnyArray
        from ..field import Field
        from ..multi_field import MultiField
        args = list(self._args)
        for i, a in enumerate(args):
            if isinstance(a, (AnyArray, Field, MultiField)):
                args[i] = a.at(x.device_id)
        self._args = tuple(args)
        for k, a in self._kwargs.items():
            if isinstance(a, (AnyArray, Field, MultiField)):
                self._kwargs[k] = a.at(x.device_id)

    def apply(self, x):
        self._check_input(x)
        self._device_preparation(x)
        return x.ptw(self._funcname, *self._args, **self._kwargs)

    def __repr__(self):
        return f"_FunctionApplier ('{self._funcname}')"


class _CombinedOperator(Operator):
    def __init__(self, ops, _callingfrommake=False):
        if not _callingfrommake:
            raise NotImplementedError
        self._ops = tuple(ops)

    @classmethod
    def unpack(cls, ops, res):
        for op in ops:
            if isinstance(op, cls):
                res = cls.unpack(op._ops, res)
            else:
                res = res + [op]
        return res

    @classmethod
    def make(cls, ops):
        res = cls.unpack(ops, [])
        if len(res) == 1:
            return res[0]
        return cls(res, _callingfrommake=True)


class _OpChain(_CombinedOperator):
    def __init__(self, ops, _callingfrommake=False):
        super(_OpChain, self).__init__(ops, _callingfrommake)
        self._domain = self._ops[-1].domain
        self._target = self._ops[0].target
        for i in range(1, len(self._ops)):
            check_object_identity(self._ops[i-1].domain, self._ops[i].target)

    def apply(self, x):
        self._check_input(x)
        for op in reversed(self._ops):
            x = op(x)
        return x

    def _simplify_for_constant_input_nontrivial(self, c_inp):
        from ..multi_domain import MultiDomain
        if not isinstance(self._domain, MultiDomain):
            return None, self
        newop = None
        for op in reversed(self._ops):
            c_inp, t_op = op.simplify_for_constant_input(c_inp)
            newop = t_op if newop is None else op(newop)
        return c_inp, newop

    def __repr__(self):
        subs = "\n".join(sub.__repr__() for sub in self._ops)
        return "_OpChain:\n" + indent(subs)


class _OpProd(Operator):
    def __init__(self, op1, op2):
        from ..sugar import domain_union
        self._domain = domain_union((op1.domain, op2.domain))
        self._target = op1.target
        if op1.target != op2.target:
            raise ValueError("target mismatch")
        self._op1 = op1
        self._op2 = op2

    def apply(self, x):
        from ..linearization import Linearization
        from ..sugar import makeOp
        self._check_input(x)
        lin = x.jac is not None
        wm = x.want_metric if lin else False
        x = x.val if lin else x
        v1 = x.extract(self._op1.domain)
        v2 = x.extract(self._op2.domain)
        if not lin:
            return self._op1(v1) * self._op2(v2)
        lin1 = self._op1(Linearization.make_var(v1, wm))
        lin2 = self._op2(Linearization.make_var(v2, wm))
        jac = (makeOp(lin1._val)(lin2._jac))._myadd(makeOp(lin2._val)(lin1._jac), False)
        return lin1.new(lin1._val*lin2._val, jac)

    def _simplify_for_constant_input_nontrivial(self, c_inp):
        from ..multi_domain import MultiDomain
        from .simplify_for_const import ConstCollector
        f1, o1 = self._op1.simplify_for_constant_input(
            c_inp.extract_part(self._op1.domain))
        f2, o2 = self._op2.simplify_for_constant_input(
            c_inp.extract_part(self._op2.domain))
        if not isinstance(self._target, MultiDomain):
            return None, _OpProd(o1, o2)
        cc = ConstCollector()
        cc.mult(f1, o1.target)
        cc.mult(f2, o2.target)
        return cc.constfield, _OpProd(o1, o2)

    def __repr__(self):
        subs = "\n".join(sub.__repr__() for sub in (self._op1, self._op2))
        return "_OpProd:\n"+indent(subs)


class _OpSum(Operator):
    def __init__(self, op1, op2):
        from ..sugar import domain_union
        self._domain = domain_union((op1.domain, op2.domain))
        self._target = domain_union((op1.target, op2.target))
        self._op1 = op1
        self._op2 = op2

    def apply(self, x):
        self._check_input(x)
        return self._apply_operator_sum(x, [self._op1, self._op2])

    @staticmethod
    def _apply_operator_sum(x, ops):
        from ..linearization import Linearization

        unite = lambda x, y: x.unite(y)
        if x.jac is None:
            return reduce(unite, (oo.force(x) for oo in ops))
        lin = [oo(Linearization.make_var(x.val.extract(oo.domain), x.want_metric))
                for oo in ops]
        jacs = map(lambda x: x._jac, lin)
        vals = map(lambda x: x._val, lin)
        metrics = list(map(lambda x: x._metric, lin))
        jac = reduce(lambda x, y: x._myadd(y, False), jacs)
        val = reduce(unite, vals)
        res = x.new(val, jac)
        if all(mm is not None for mm in metrics):
            res = res.add_metric(reduce(add, metrics))
        return res

    def _simplify_for_constant_input_nontrivial(self, c_inp):
        from ..multi_domain import MultiDomain
        from .simplify_for_const import ConstCollector
        f1, o1 = self._op1.simplify_for_constant_input(
            c_inp.extract_part(self._op1.domain))
        f2, o2 = self._op2.simplify_for_constant_input(
            c_inp.extract_part(self._op2.domain))
        if not isinstance(self._target, MultiDomain):
            return None, _OpSum(o1, o2)
        cc = ConstCollector()
        cc.add(f1, o1.target)
        cc.add(f2, o2.target)
        return cc.constfield, _OpSum(o1, o2)

    def __repr__(self):
        subs = "\n".join(sub.__repr__() for sub in (self._op1, self._op2))
        return "_OpSum:\n"+indent(subs)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2021 Max-Planck-Society
# Copyright(C) 2025 Philipp Arras
#
# Author: Philipp Arras, Philipp Frank

import os
import pathlib
import pickle
import re
import time
from warnings import warn

import numpy as np

from .. import utilities
from ..field import Field
from ..multi_domain import MultiDomain
from ..multi_field import MultiField
from ..operators.operator import Operator
from ..utilities import (ensure_all_tasks_succeed, get_MPI_params_from_comm,
                         shareRange)


class SampleListBase:
    """Base class for storing lists of fields representing samples.

    This class suits as a base class for storing lists of in most cases
    posterior samples. It is intended to be used to hold the minimization state
    of an inference run and comes with a variety of convenience functions like
    computing the mean or standard deviation of the output of a given operator
    over the sample list.

    Parameters
    ----------
    comm : MPI communicator or None
        If not `None`, :class:`SampleListBase` can gather samples across multiple
        MPI tasks. If `None`, :class:`SampleListBase` is not a distributed object.
    domain : Domainoid (can be DomainTuple, MultiDomain, dict, Domain or list of Domains)
        The domain on which the samples are defined.
    n_local_samples : int
        Number of samples present on the current MPI task.

    Note
    ----
    A class inheriting from :class:`SampleListBase` needs to call the constructor of
    `SampleListBase` and needs to implement :attr:`n_local_samples` and :attr:`local_item()`.
    """
    def __init__(self, comm, domain, n_local_samples):
        from ..sugar import makeDomain
        self._comm = comm
        self._domain = makeDomain(domain)
        utilities.check_MPI_equality(self._domain, comm)
        self.local_indices = _compute_local_indices(n_local_samples, comm)

        self._active_comm = comm
        if comm is not None:
            # 0 corresponds to empty tasks
            # 1 corresponds to tasks with samples
            color = 0 if len(self.local_indices) == 0 else 1
            # If the same key is assigned to multiple processes, the conflict is
            # resolved according to the rank in the original communicator
            key = 0
            self._active_comm = comm.Split(color, key)

        self._n_samples = utilities.allreduce_sum([self.n_local_samples], self.comm)

    @property
    def n_local_samples(self):
        """int: Number of local samples."""
        return len(self.local_indices)

    @property
    def n_samples(self):
        """Return number of samples across all MPI tasks."""
        return self._n_samples

    def local_item(self, i):
        """Return ith local sample."""
        raise NotImplementedError

    def local_iterator(self):
        """Return an iterator over all local samples."""
        for i in range(self.n_local_samples):
            yield self.local_item(i)

    @property
    def comm(self):
        """MPI communicator or None: The communicator used for the SampleListBase."""
        return self._comm

    @property
    def MPI_master(self):
        return utilities.get_MPI_params_from_comm(self.comm)[2]

    @property
    def domain(self):
        """DomainTuple or MultiDomain: the domain on which the samples are defined."""
        return self._domain

    def save_to_hdf5(self, file_name, op=None, samples=False, mean=False, std=False,
                     overwrite=False):
        """Write sample list to HDF5 file.

        This function writes sample lists to HDF5 files that contain two
        groups: `samples` and `stats`. `samples` contain the sublabels `0`,
        `1`, ... that number the labels and `stats` contains the sublabels
        `mean` and `standard deviation`. If `self.domain` is an instance of
        :class:`~nifty.cl.multi_domain.MultiDomain`, these sublabels refer
        themselves to subgroups. For :class:`~nifty.cl.field.Field`, the sublabel
        refers to an HDF5 data set.

        If quanitities are not requested (e.g. by setting `mean=False`), the
        respective sublabels are not present in the HDF5 file.

        If `op` is an :class:`~nifty.cl.operators.operator.Operator`, the operator
        string representation, its domain and its target are written to the
        attributes of the HDF5 file.

        Parameters
        ----------
        file_name : str
            File name of output hdf5 file.
        op : callable or None
            Callable that is applied to each item in the :class:`SampleListBase`
            before it is returned. Can be an
            :class:`~nifty.cl.operators.operator.Operator` or any other callable
            that takes a :class:`~nifty.cl.field.Field` as an input. Default:
            None.
        samples : bool
            If True, samples are written into hdf5 file.
        mean : bool
            If True, mean of samples is written into hdf5 file.
        std : bool
            If True, standard deviation of samples is written into hdf5 file.
        overwrite : bool
            If True, a potentially existing file with the same file name as
            `file_name`, is overwritten.
        """
        import h5py

        if os.path.isfile(file_name):
            if self.MPI_master and overwrite:
                os.remove(file_name)
            if not overwrite:
                raise RuntimeError(f"File {file_name} already exists. Delete it or use "
                                   "`overwrite=True`")
        if not (samples or mean or std):
            raise ValueError("Neither samples nor mean nor standard deviation shall be written.")

        if self.MPI_master:
            f = h5py.File(file_name, "w")
            if isinstance(op, Operator):
                f.attrs["nifty operator string representation"] = str(op)
                f.attrs["nifty operator domain"] = repr(op.domain)
                f.attrs["nifty operator target"] = repr(op.target)
                f.attrs["nifty domain"] = repr(op.target)
            else:
                f.attrs["nifty domain"] = repr(self.domain)
        else:
            f = utilities.Nop()

        # TODO Add some meta information (e.g. k_length for PowerSpace, distances for RGSpace)
        if samples:
            grp = f.create_group("samples")
            for ii, ss in enumerate(self.iterator(op)):
                _field2hdf5(grp, ss, str(ii))
        if std:
            grp = f.create_group("stats")
            m, v = self.sample_stat(op)
        else:
            if mean:
                grp = f.create_group("stats")
                m = self.average(op)
        if mean:
            _field2hdf5(grp, m, "mean")
        if std:
            _field2hdf5(grp, v.sqrt(), "standard deviation")

        f.close()
        _barrier(self.comm)

    def save_to_fits(self, file_name_base, op=None, samples=False, mean=False, std=False,
                     overwrite=False):
        """Write sample list to FITS file.

        This function writes properties of a sample list to a FITS file. This is
        supported if and only if the target of `op` is a two-dimensional
        RGSpace.

        Parameters
        ----------
        file_name_base : str
            File name base of output FITS file, i.e. without `.fits` extension.
        op : callable or None
            Callable that is applied to each item in the :class:`SampleListBase`
            before it is returned. Can be an
            :class:`~nifty.cl.operators.operator.Operator` or any other callable
            that takes a :class:`~nifty.cl.field.Field` as an input. Default:
            None.
        samples : bool
            If True, samples are written into hdf5 file.
        mean : bool
            If True, mean of samples is written into hdf5 file.
        std : bool
            If True, standard deviation of samples is written into hdf5 file.
        overwrite : bool
            If True, a potentially existing file with the same file name as
            `file_name`, is overwritten.
        """
        # TODO Add support for 3d and 4d fields
        # https://fits.gsfc.nasa.gov/fits_standard.html

        if not (samples or mean or std):
            raise ValueError("Neither samples nor mean nor standard deviation shall be written.")

        if mean or std:
            m, v = self.sample_stat(op)
        if mean:
            self._save_fits_2d(m, file_name_base + "_mean.fits", overwrite)
        if std:
            self._save_fits_2d(v.sqrt(), file_name_base + "_std.fits", overwrite)

        if samples:
            for ii, ss in enumerate(self.iterator(op)):
                self._save_fits_2d(ss, file_name_base + f"_sample_{ii}.fits", overwrite)

    def _save_fits_2d(self, fld, file_name, overwrite):
        import astropy.io.fits as pyfits
        from astropy.time import Time

        from ..domain_tuple import DomainTuple

        dom = fld.domain
        if not isinstance(dom, DomainTuple) or len(dom) != 1 or len(dom[0].shape) != 2:
            raise ValueError("FITS file export is only supported from 2d-fields. "
                             f"Current domain:\n{dom}")
        h = pyfits.Header()
        h["DATE-MAP"] = Time(time.time(), format="unix").iso.split()[0]
        h["CRVAL1"] = h["CRVAL2"] = 0
        h["CRPIX1"] = h["CRPIX2"] = 0
        h["CUNIT1"] = h["CUNIT2"] = "deg"
        h["CDELT1"], h["CDELT2"] = -dom[0].distances[0], dom[0].distances[1]
        h["CTYPE1"] = "RA---SIN"
        h["CTYPE2"] = "DEC---SIN"
        h["EQUINOX"] = 2000

        hdu = pyfits.PrimaryHDU(fld.asnumpy()[:, :].T, header=h)
        hdulist = pyfits.HDUList([hdu])
        if self.MPI_master:
            hdulist.writeto(file_name, overwrite=overwrite)

    def iterator(self, op=None):
        """Return iterator over all potentially distributed samples.

        Parameters
        ----------
        op : callable or None
            Callable that is applied to each item in the :class:`SampleListBase`
            before it is returned. Can be an
            :class:`~nifty.cl.operators.operator.Operator` or any other callable
            that takes a :class:`~nifty.cl.field.Field` as an input. Default:
            None.

        Note
        ----
        Calling this function involves MPI communication if `comm != None`.
        """
        op = _none_to_id(op)
        if self.comm is not None:
            for itask in range(self.comm.Get_size()):
                for i in range(_bcast(self.n_local_samples, self._comm, itask)):
                    ss = self.local_item(i) if itask == self._comm.Get_rank() else None
                    yield op(_bcast(ss, self._comm, itask))
        else:
            for ss in self.local_iterator():
                yield op(ss)

    def average(self, op=None):
        """Compute average over all potentially distributed samples.

        Parameters
        ----------
        op : callable or None
            Callable that is applied to each item in the :class:`SampleListBase`
            before it is averaged.


        Note
        ----
        Calling this function involves MPI communication if `comm != None`.

        Note
        ----
        Calling this function involves allocating arrays for all samples and
        averaging them afterwards. If the number of local samples is big and
        `op` is not None, this leads to much temporary memory usage. If the
        output of `op` is just a :class:`~nifty.cl.field.Field` or
        :class:`~nifty.cl.multi_field.MultiField`, :attr:`sample_stat()`
        can be used in order to compute the average memory efficiently.
        """
        res = self._prepare_average(op)
        n = self.n_samples
        return utilities.allreduce_sum(res, self.comm) / n

    def _average_tuple(self, op):
        """Compute simultaneous average over all potentially distributed samples.

        Parameters
        ----------
        op : callable
            Callable that is applied to each item in the
            :class:`SampleListBase` before it is averaged. `op` shall return a
            tuple of whose individual averages are computed and returned as
            tuple.

        Note
        ----
        Calling this function involves MPI communication if `comm != None`.
        """
        n = self.n_samples
        res = None
        if len(self.local_indices) > 0:
            res = self._prepare_average(op)
            res = _list_transpose(res)
            res = tuple(utilities.allreduce_sum(rr, self._active_comm) / n for rr in res)
        if self._active_comm is None and self._comm is None:
            return res

        if len(self.local_indices) > 0:
            task_list = self._comm.allgather(self._comm.Get_rank())
        else:
            task_list = self._comm.allgather(-1)
        root = next(filter((-1).__ne__, task_list))
        return self._comm.bcast(res, root=root)

    def _prepare_average(self, op):
        op = _none_to_id(op)
        res = [op(ss) for ss in self.local_iterator()]
        return res

    def sample_stat(self, op=None):
        """Compute mean and variance of samples after applying `op`.

        Parameters
        ----------
        op : callable or None
            Callable that is applied to each item in the :class:`SampleListBase`
            before it is used to compute mean and variance.

        Returns
        -------
        tuple
            A tuple with two items: the mean and the variance.
        """
        from ..probing import StatCalculator
        if self.n_samples == 1:
            res = self.average(op)
            return res, 0*res
        sc = StatCalculator()
        for ss in self.iterator(op):
            sc.add(ss)
        return sc.mean, sc.var

    def save(self, file_name_base, overwrite=False):
        """Serialize SampleList and write it to disk.

        Parameters
        ----------
        file_name_base : str
            File name of the output file without extension. The actual file name
            will have the extension ".pickle".
        overwrite : bool
            Existing files are overwritten.

        Note
        ----
        If the instance of :class:`SampleListBase` is distributed, each MPI task
        writes its own file.
        """
        raise NotImplementedError

    @classmethod
    def load(cls, file_name_base, comm=None):
        """Deserialize SampleList from files on disk.

        Parameters
        ----------
        file_name_base : str
            File name of the input file without extension. The actual file name
            will have the extension ".pickle".
        comm : MPI communicator or None
            If not `None`, each MPI task reads its own input file.

        Note
        ----
        `file_name_base` needs to be the same string that has been used for
        saving the :class:`SampleListBase`.
        """
        raise NotImplementedError

    @classmethod
    def _list_local_sample_files(cls, file_name_base, comm=None):
        """List all sample files that are relevant for the local task.

        All sample files that correspond to `file_name_base` are searched and
        selected based on which rank the local task has.

        Note
        ----
        This function makes sure that the all file numbers between 0 and the
        maximal found number are present. If this is not the case, a
        `RuntimeError` is raised.

        Note
        ----
        This function distributes the samples according to the standard NIFTy
        distribution scheme (see `ift.utilities.shareRange`).
        """
        base_dir, base_file = os.path.split(os.path.abspath(file_name_base))
        files = [ff for ff in os.listdir(base_dir)
                 if re.match(f"{base_file}.[0-9]+.pickle", ff)]
        if len(files) == 0:
            raise RuntimeError(f"No files matching `{file_name_base}.*.pickle`")
        n_samples = _consecutive_length(list(map(lambda x: int(x.split(".")[-2]), files)))

        ntask, rank, _ = get_MPI_params_from_comm(comm)
        local_indices = range(*shareRange(n_samples, ntask, rank))
        files = [f"{file_name_base}.{ii}.pickle" for ii in local_indices]
        for ff in files:
            if not os.path.isfile(ff):
                raise RuntimeError(f"File {ff} not found")
        return files

    def deep_eq(self, other):
        # TODO: Potentially add __eq__ to MultiField and Field to simplify this
        if not isinstance(other, SampleListBase):
            return NotImplemented
        if self.n_samples != other.n_samples:
            return False
        for xx, yy in zip(self.iterator(), other.iterator()):
            if xx.domain is not yy.domain:
                return False
            if isinstance(xx, MultiField):
                xx = list(xx.val.values())
                yy = list(yy.val.values())
            else:
                xx, yy = [xx.val], [yy.val]
            assert len(xx) == len(yy)
            for xxx, yyy in zip(xx, yy):
                if np.any(xxx != yyy):
                    return False
        return True


class ResidualSampleList(SampleListBase):
    def __init__(self, mean, residuals, neg, comm=None):
        """Store samples in terms of a mean and a residual deviation thereof.


        Parameters
        ----------
        mean : :class:`nifty.cl.field.Field` or :class:`nifty.cl.multi_field.MultiField`
            Mean of the sample list.
        residuals : list of :class:`nifty.cl.field.Field` or list of :class:`nifty.cl.multi_field.MultiField`
            List of residuals from the mean. If it is a list of `MultiField`,
            the domain of the residuals can be a subdomain of the domain of
            mean. This results in adding just a zero in respective `MultiField`
            entries.
        neg: list of bool
            This list has to have the same length as `residuals`. If an entry is
            `True`, the respective residual is subtracted and not added.
        comm : MPI communicator or None
            If not `None`, samples can be gathered across multiple MPI tasks. If
            `None`, :class:`ResidualSampleList` is not a distributed object.
        """
        self._m = mean
        self._r = tuple(residuals)
        self._n = tuple(neg)
        super(ResidualSampleList, self).__init__(comm, mean.domain, len(self._r))

        if len(self._r) != len(self._n):
            raise ValueError("Residuals and neg need to have the same length.")
        if any(elem == 0 for elem in utilities.allreduce_sum([[len(self._r)]], comm)):
            warn("ResidualSampleList: (Partially empty) sample list detected.")

        if len(self._r) > 0:
            r_dom = self._r[0].domain
            if not all(rr.domain is r_dom for rr in self._r):
                raise ValueError("All residuals must have the same domain.")
            if isinstance(r_dom, MultiDomain):
                try:
                    self._m.extract(r_dom)
                except:
                    raise ValueError("`residual.domain` must be a subdomain of `mean.domain`.")

        if not all(isinstance(nn, bool) for nn in neg):
            raise TypeError("All entries in neg need to be bool.")

    def local_item(self, i):
        return self._m.flexible_addsub(self._r[i], self._n[i])

    def at(self, mean):
        """Instantiate `ResidualSampleList` with the same residuals as `self`.

        The mean is updated.

        Note
        ----
        If `self.domain` is a :class:`~nifty.cl.multi_domain.MultiDomain`, the old
        and new mean are combined with
        :attr:`~nifty.cl.multi_field.MultiField.union` beforehand. This means that
        only the multi field entries present in `mean` are updated.

        Returns
        -------
        ResidualSampleList
            Sample list with updated mean.
        """
        if isinstance(self._m, MultiField) and self.domain is not mean.domain:
            mean = MultiField.union([self._m, mean])
        return ResidualSampleList(mean, self._r, self._n, self.comm)

    @property
    def mean(self):
        """Read-only view on the `mean` of the `ResidualSampleList`.

        Note
        ----
        This mean should not be used to compute the sample average of a
        non-linear observable/ operator (use `.average(operator)` instead).
        """
        return self._m

    def save(self, file_name_base, overwrite=False):
        # TODO: Make this function atomic potentially unify with
        #  SampleList.save.

        _ensure_proper_sample_list_ending(_sample_file_name(file_name_base, self.n_samples),
                                          overwrite, self.comm)

        # Save samples
        with ensure_all_tasks_succeed(self.comm):
            for ii, isample in enumerate(self.local_indices):
                obj = [self._r[ii], self._n[ii]]
                fname = _sample_file_name(file_name_base, isample)
                _save_to_disk(fname, obj, overwrite)

        # Save mean
        with ensure_all_tasks_succeed(self.comm):
            if self.MPI_master:
                _save_to_disk(f"{file_name_base}.mean.pickle", self._m, overwrite)

    @classmethod
    def load(cls, file_name_base, comm=None):
        _barrier(comm)
        mean = _load_from_disk(f"{file_name_base}.mean.pickle")
        files = cls._list_local_sample_files(file_name_base, comm)
        tmp = [_load_from_disk(ff) for ff in files]
        res = [aa[0] for aa in tmp]
        neg = [aa[1] for aa in tmp]
        return cls(mean, res, neg, comm=comm)

    @classmethod
    def load_mean(cls, file_name_base):
        return _load_from_disk(f"{file_name_base}.mean.pickle")


class SampleList(SampleListBase):
    def __init__(self, samples, comm=None, domain=None):
        """Store samples as a plain list.

        This is a minimalist implementation of :class:`SampleListBase`. It just
        serves as a (potentially MPI-distributed) wrapper of a list of samples.

        Parameters
        ----------
        samples : list of :class:`nifty.cl.field.Field` or list of :class:`nifty.cl.multi_field.MultiField`
            List of samples.
        comm : MPI communicator or None
            If not `None`, samples can be gathered across multiple MPI tasks. If
            `None`, :class:`ResidualSampleList` is not a distributed object.
        domain : DomainTuple, MultiDomain or None
            Sets the domain of the `SampleList`. If `samples` is non-empty and
            `domain` is not None, `domain` has to coincide with the domain of
            the samples. Default: None.
        """
        from ..sugar import makeDomain

        if domain is None:
            if len(samples) == 0:
                raise ValueError("Need to pass `domain` to instantiate empty `SampleList`.")
            else:
                domain = samples[0].domain
        else:
            domain = makeDomain(domain)
        super(SampleList, self).__init__(comm, domain, len(samples))
        self._s = samples
        for ss in self._s:
            utilities.check_object_identity(ss.domain, self.domain)

    def local_item(self, i):
        return self._s[i]

    def save(self, file_name_base, overwrite=False):
        # TODO: Make this function atomic potentially unify with
        # ResidualSampleList.save.

        _ensure_proper_sample_list_ending(_sample_file_name(file_name_base, self.n_samples),
                                          overwrite, self.comm)

        # Save samples
        with ensure_all_tasks_succeed(self.comm):
            for ii, isample in enumerate(self.local_indices):
                obj = self._s[ii]
                fname = _sample_file_name(file_name_base, isample)
                _save_to_disk(fname, obj, overwrite)

    @classmethod
    def load(cls, file_name_base, comm=None):
        from ..logger import logger
        _barrier(comm)
        foo = "{file_name_base}.mean.pickle"
        if os.path.isfile(foo):
            logger.warning(f"{foo} is present. Most probably you intended to "
                         "call `ift.ResidualSampleList.load()`.")
        files = cls._list_local_sample_files(file_name_base, comm)
        samples = [_load_from_disk(ff) for ff in files]
        dom = None
        if comm is not None:
            if comm.Get_rank() == 0:
                dom = samples[0].domain
            dom = comm.bcast(dom, 0)
        return cls(samples, comm=comm, domain=dom)


def _none_to_id(obj):
    """If the input is None, replace it with identity map. If input is
    `Operator`, append `force` in order to make sure that the function is
    evaluated if there is a chance that the domains match. Otherwise return
    input.
    """
    if obj is None:
        return lambda x: x
    if isinstance(obj, Operator):
        return lambda x: obj.force(x)
    return obj


def _bcast(obj, comm, root):
    """Broadcast python object from given root.

    Parameters
    ----------
    obj : object
        The object to be broadcasted.
    comm : MPI communicator
        MPI communicator used for the broadcasting.
    root : int
        MPI task number from which the object shall be sent.
    """
    data = obj if comm.Get_rank() == root else None
    return comm.bcast(data, root=root)


def _sample_file_name(file_name_base, isample):
    """Return sample-unique file name.

    This file name that can be used to uniquely write samples potentially from all MPI tasks.

    Parameters
    ----------
    file_name_base : str
        User-defined first part of file name.
    isample : int
        Number of sample
    """
    if not isinstance(isample, int):
        raise TypeError
    return f"{file_name_base}.{isample}.pickle"


def _load_from_disk(file_name):
    with open(file_name, "rb") as f:
        obj = pickle.load(f)
    return obj


def _save_to_disk(file_name, obj, overwrite=False):
    if not overwrite and os.path.isfile(file_name):
        raise RuntimeError(f"{file_name} already exists")
    if overwrite and os.path.isfile(file_name):
        os.remove(file_name)
    with open(file_name, "wb") as f:
        pickle.dump(obj, f, pickle.HIGHEST_PROTOCOL)


def _field2hdf5(file_handle, obj, name):
    if not isinstance(name, str):
        raise TypeError
    if isinstance(obj, MultiField):
        grp = file_handle.create_group(name)
        for kk, fld in obj.items():
            _field2hdf5(grp, fld, kk)
        return
    if not isinstance(obj, Field):
        raise TypeError
    file_handle.create_dataset(name, data=obj.asnumpy())


def _barrier(comm):
    if comm is not None:
        comm.Barrier()


def _list_transpose(list_of_lists):
    ny = len(list_of_lists[0])
    return [[elem[ii] for elem in list_of_lists] for ii in range(ny)]


def _compute_local_indices(n_local, comm):
    if comm is None:
        return range(n_local)
    n_locals = comm.allgather(n_local)
    start = sum(n_locals[:comm.Get_rank()])
    return range(start, start + n_local)


def _consecutive_length(lst):
    if 0 not in lst:
        raise ValueError("List does not contain 0, the starting element")
    res = 0
    while True:
        if res + 1 not in lst:
            return res + 1
        res += 1


def _ensure_proper_sample_list_ending(fname, overwrite, comm):
        with ensure_all_tasks_succeed(comm):
            MPI_master = utilities.get_MPI_params_from_comm(comm)[2]
            if MPI_master:
                if overwrite:
                    # Remove potential "next sample"

                    # TODO: this might leave dangling samples that are either
                    # ignored or overwritten later. So it is fine for now. Will
                    # be deleted once a proper atomic approach is implemented.
                    pathlib.Path(fname).unlink(missing_ok=True)
                else:
                    # Make sure that "next" sample does not exist
                    if os.path.isfile(fname):
                        raise RuntimeError(f"{fname} already exists. You may "
                            "want to remove it or specify overwrite=True")



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2021 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

from .endomorphic_operator import EndomorphicOperator
from .linear_operator import LinearOperator
from .scaling_operator import ScalingOperator


class SandwichOperator(EndomorphicOperator):
    """Operator which is equivalent to the expression
    `bun.adjoint(cheese(bun))`.

    Note
    ----
    This operator should always be called using the `make` method.
    """

    def __init__(self, bun, cheese, op, _callingfrommake=False):
        if not _callingfrommake:
            raise NotImplementedError
        self._bun = bun
        self._cheese = cheese
        self._op = op
        self._domain = op.domain
        self._capability = op._capability

    @staticmethod
    def make(bun, cheese=None, sampling_dtype=None):
        """Build a SandwichOperator (or something simpler if possible)

        Parameters
        ----------
        bun: LinearOperator
            the bun part
        cheese: EndomorphicOperator
            the cheese part
        sampling_dtype :
            If this operator represents the covariance of a Gaussian probabilty
            distribution and cheese is `None`, `sampling_dtype` specifies if it
            is real or complex Gaussian. If `sampling_dtype` and `cheese` are
            `None`, the operator cannot be used as a covariance, i.e. no samples
            can be drawn. Default: None.
        """
        if isinstance(cheese, SandwichOperator):
            old_cheese = cheese
            cheese = old_cheese._cheese
            bun = old_cheese._bun @ bun

        if not isinstance(bun, LinearOperator):
            raise TypeError("bun must be a linear operator")
        if cheese is not None and not isinstance(cheese, LinearOperator):
            raise TypeError("cheese must be a linear operator or None")
        if cheese is None:
            cheese = ScalingOperator(bun.target, 1., sampling_dtype)
        if isinstance(bun, ScalingOperator):
            fct = abs(bun._factor)**2
            if fct == 1.:
                return cheese
            op = cheese.scale(fct)
        else:
            op = bun.adjoint @ cheese @ bun
        return SandwichOperator(bun, cheese, op, _callingfrommake=True)

    def apply(self, x, mode):
        return self._op.apply(x, mode)

    def draw_sample(self, from_inverse=False, device_id=-1):
        # Inverse samples from general sandwiches are not possible
        if from_inverse:
            if self._bun.capability & self._bun.INVERSE_TIMES:
                try:
                    s = self._cheese.draw_sample(from_inverse, device_id)
                    return self._bun.inverse_times(s)
                except NotImplementedError:
                    pass
            raise NotImplementedError(
                "cannot draw from inverse of this operator")

        # Samples from general sandwiches
        return self._bun.adjoint_times(
            self._cheese.draw_sample(from_inverse, device_id))

    def get_sqrt(self):
        if self._cheese is None:
            return self._bun
        return self._cheese.get_sqrt() @ self._bun

    def __repr__(self):
        from ..utilities import indent
        return "\n".join((
            "SandwichOperator:",
            indent("\n".join((
                "Cheese:", self._cheese.__repr__(),
                "Bun:", self._bun.__repr__())))))



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2019 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import numpy as np

from ..logger import logger
from ..probing import approximation2endo
from ..sugar import makeOp
from .conjugate_gradient import ConjugateGradient
from .iteration_controllers import (AbsDeltaEnergyController,
                                    GradientNormController)
from .line_search import LineSearch
from .minimizer import Minimizer
from .quadratic_energy import QuadraticEnergy


class DescentMinimizer(Minimizer):
    """A base class used by gradient methods to find a local minimum.

    Descent minimization methods are used to find a local minimum of a scalar
    function by following a descent direction. This class implements the
    minimization procedure once a descent direction is known. The descent
    direction has to be implemented separately.

    Parameters
    ----------
    controller : IterationController
        Object that decides when to terminate the minimization.
    line_searcher : callable *optional*
        Function which infers the step size in the descent direction
        (default : LineSearch()).
    """

    def __init__(self, controller, line_searcher=LineSearch()):
        self._controller = controller
        self.line_searcher = line_searcher

    def __call__(self, energy):
        """Performs the minimization of the provided Energy functional.

        Parameters
        ----------
        energy : Energy
           Energy object which provides value, gradient and metric at a
           specific position in parameter space.

        Returns
        -------
        Energy
            Latest `energy` of the minimization.
        int
            Can be controller.CONVERGED or controller.ERROR

        Notes
        -----
        The minimization is stopped if
            * the controller returns controller.CONVERGED or controller.ERROR,
            * a perfectly flat point is reached,
            * according to the line-search the minimum is found,
        """
        f_k_minus_1 = None
        controller = self._controller
        status = controller.start(energy)
        if status != controller.CONTINUE:
            return energy, status

        while True:
            # check if position is at a flat point
            if energy.gradient_norm == 0:
                return energy, controller.CONVERGED

            # compute a step length that reduces energy.value sufficiently
            new_energy, success = self.line_searcher.perform_line_search(
                energy=energy,
                pk=self.get_descent_direction(energy, f_k_minus_1),
                f_k_minus_1=f_k_minus_1)
            if not success:
                self.reset()

            f_k_minus_1 = energy.value

            if new_energy.value > energy.value:
                logger.error("Error: Energy has increased")
                return energy, controller.ERROR

            if new_energy.value == energy.value:
                logger.warning(
                    "Warning: Energy has not changed. Assuming convergence...")
                return new_energy, controller.CONVERGED

            energy = new_energy
            status = self._controller.check(energy)
            if status != controller.CONTINUE:
                return energy, status

    def reset(self):
        pass

    def get_descent_direction(self, energy, old_value=None):
        """Calculates the next descent direction.

        Parameters
        ----------
        energy : Energy
            An instance of the Energy class which shall be minimized. The
            position of `energy` is used as the starting point of minimization.

        old_value : float
            if provided, this must be the value of the energy in the previous
            step.

        Returns
        -------
        :class:`nifty.cl.field.Field` or :class:`nifty.cl.multi_field.MultiField`
           The descent direction.
        """
        raise NotImplementedError

    @property
    def controller(self):
        return self._controller


class SteepestDescent(DescentMinimizer):
    """Implementation of the steepest descent minimization scheme.

    Also known as 'gradient descent'. This algorithm simply follows the
    functional's gradient for minimization.
    """

    def get_descent_direction(self, energy, _=None):
        return -energy.gradient


class RelaxedNewton(DescentMinimizer):
    """Calculates the descent direction according to a Newton scheme.

    The descent direction is determined by weighting the gradient at the
    current parameter position with the inverse local metric.
    """

    def __init__(self, controller, line_searcher=None):
        if line_searcher is None:
            line_searcher = LineSearch(preferred_initial_step_size=1.)
        super(RelaxedNewton, self).__init__(controller=controller,
                                            line_searcher=line_searcher)

    def get_descent_direction(self, energy, _=None):
        return -energy.metric.inverse_times(energy.gradient)


class NewtonCG(DescentMinimizer):
    """Calculates the descent direction according to a Newton-CG scheme.

    Algorithm derived from SciPy sources.
    """

    def __init__(self, controller, napprox=0, line_searcher=None, name=None,
                 nreset=20, max_cg_iterations=200, energy_reduction_factor=0.1,
                 enable_logging=False):
        if line_searcher is None:
            line_searcher = LineSearch(preferred_initial_step_size=1.)
        super(NewtonCG, self).__init__(controller=controller,
                                       line_searcher=line_searcher)
        self._napprox = napprox
        self._name = name
        self._nreset = nreset
        self._max_cg_iterations = max_cg_iterations
        self._alpha = energy_reduction_factor
        from .iteration_controllers import EnergyHistory
        self._history = EnergyHistory() if enable_logging else None

    def get_descent_direction(self, energy, old_value=None):
        if old_value is None:
            ic = GradientNormController(iteration_limit=5)
        else:
            ediff = self._alpha*(old_value-energy.value)
            ic = AbsDeltaEnergyController(
                ediff, iteration_limit=self._max_cg_iterations, name=self._name)
        if self._history is not None:
            ic.enable_logging()
        e = QuadraticEnergy(0*energy.position, energy.metric, energy.gradient)
        p = None
        if self._napprox > 1:
            met = energy.metric
            p = makeOp(approximation2endo(met, self._napprox)).inverse
        e, conv = ConjugateGradient(ic, nreset=self._nreset)(e, p)
        if self._history is not None:
            self._history += ic.history
        if conv == ic.ERROR:
            raise ValueError("Cannot find descent direction")
        return -e.position

    @property
    def inversion_history(self):
        return self._history


class L_BFGS(DescentMinimizer):
    def __init__(self, controller, line_searcher=LineSearch(),
                 max_history_length=5):
        super(L_BFGS, self).__init__(controller=controller,
                                     line_searcher=line_searcher)
        self.max_history_length = max_history_length

    def __call__(self, energy):
        self.reset()
        return super(L_BFGS, self).__call__(energy)

    def reset(self):
        self._k = 0
        self._s = [None]*self.max_history_length
        self._y = [None]*self.max_history_length

    def get_descent_direction(self, energy, _=None):
        x = energy.position
        s = self._s
        y = self._y
        k = self._k
        maxhist = self.max_history_length
        gradient = energy.gradient

        nhist = min(k, maxhist)
        alpha = [None]*maxhist
        p = -gradient
        if k > 0:
            idx = (k-1) % maxhist
            s[idx] = x-self._lastx
            y[idx] = gradient-self._lastgrad
        if nhist > 0:
            for i in range(k-1, k-nhist-1, -1):
                idx = i % maxhist
                alpha[idx] = s[idx].s_vdot(p)/s[idx].s_vdot(y[idx])
                p = p - alpha[idx]*y[idx]
            idx = (k-1) % maxhist
            fact = s[idx].s_vdot(y[idx]) / y[idx].s_vdot(y[idx])
            if fact <= 0.:
                logger.error("L-BFGS curvature not positive definite!")
            p = p*fact
            for i in range(k-nhist, k):
                idx = i % maxhist
                beta = y[idx].s_vdot(p) / s[idx].s_vdot(y[idx])
                p = p + (alpha[idx]-beta)*s[idx]
        self._lastx = x
        self._lastgrad = gradient
        self._k += 1
        return p


class VL_BFGS(DescentMinimizer):
    """Implementation of the Vector-free L-BFGS minimization scheme.

    Find the descent direction by using the inverse Hessian.
    Instead of storing the whole matrix, it stores only the last few
    updates, which are used to do operations requiring the inverse
    Hessian product. The updates are represented in a new basis to optimize
    the algorithm.

    References
    ----------
    W. Chen, Z. Wang, J. Zhou, "Large-scale L-BFGS using MapReduce", 2014,
    Microsoft
    """

    def __init__(self, controller, line_searcher=LineSearch(),
                 max_history_length=5):
        super(VL_BFGS, self).__init__(controller=controller,
                                      line_searcher=line_searcher)
        self.max_history_length = max_history_length

    def __call__(self, energy):
        self._information_store = None
        return super(VL_BFGS, self).__call__(energy)

    def reset(self):
        self._information_store = None

    def get_descent_direction(self, energy, _=None):
        x = energy.position
        gradient = energy.gradient
        # initialize the information store if it doesn't already exist
        try:
            self._information_store.add_new_point(x, gradient)
        except AttributeError:
            self._information_store = _InformationStore(
                self.max_history_length, x0=x, gradient=gradient)

        b = self._information_store.b
        delta = self._information_store.delta

        descent_direction = delta[0] * b[0]
        for i in range(1, len(delta)):
            descent_direction = descent_direction + delta[i]*b[i]

        return descent_direction


class _InformationStore:
    """Class for storing a list of past updates.

    Parameters
    ----------
    max_history_length : int
        Maximum number of stored past updates.
    x0 : :class:`nifty.cl.field.Field`
        Initial position in variable space.
    gradient : :class:`nifty.cl.field.Field`
        Gradient at position x0.

    Attributes
    ----------
    max_history_length : int
        Maximum number of stored past updates.
    s : List
        Circular buffer of past position differences, which are Fields.
    y : List
        Circular buffer of past gradient differences, which are Fields.
    last_x : :class:`nifty.cl.field.Field`
        Latest position in variable space.
    last_gradient : :class:`nifty.cl.field.Field`
        Gradient at latest position.
    k : int
        Number of updates that have taken place
    ss : numpy.ndarray
        2D circular buffer of scalar products between different elements of s.
    sy : numpy.ndarray
        2D circular buffer of scalar products between elements of s and y.
    yy : numpy.ndarray
        2D circular buffer of scalar products between different elements of y.
    """

    def __init__(self, max_history_length, x0, gradient):
        self.max_history_length = max_history_length
        self.s = [None]*max_history_length
        self.y = [None]*max_history_length
        self.last_x = x0
        self.last_gradient = gradient
        self.k = 0

        mmax = max_history_length
        self.ss = np.empty((mmax, mmax), dtype=np.float64)
        self.sy = np.empty((mmax, mmax), dtype=np.float64)
        self.yy = np.empty((mmax, mmax), dtype=np.float64)

    @property
    def history_length(self):
        """Returns the number of currently stored updates."""
        return min(self.k, self.max_history_length)

    @property
    def b(self):
        """Combines s, y and gradient to form the new base vectors b.

        Returns
        -------
        List
            List of new basis vectors.
        """
        result = []
        m = self.history_length
        mmax = self.max_history_length

        for i in range(m):
            result.append(self.s[(self.k-m+i) % mmax])

        for i in range(m):
            result.append(self.y[(self.k-m+i) % mmax])

        result.append(self.last_gradient)

        return result

    @property
    def b_dot_b(self):
        """Generates the (2m+1) * (2m+1) scalar matrix.

        The i,j-th element of the matrix is a scalar product between the i-th
        and j-th base vector.

        Returns
        -------
        numpy.ndarray
            Scalar matrix.
        """
        m = self.history_length
        mmax = self.max_history_length
        k = self.k
        result = np.empty((2*m+1, 2*m+1), dtype=np.float64)

        # update the stores
        k1 = (k-1) % mmax
        for i in range(m):
            kmi = (k-m+i) % mmax
            self.ss[kmi, k1] = self.ss[k1, kmi] = self.s[kmi].s_vdot(self.s[k1])
            self.yy[kmi, k1] = self.yy[k1, kmi] = self.y[kmi].s_vdot(self.y[k1])
            self.sy[kmi, k1] = self.s[kmi].s_vdot(self.y[k1])
        for j in range(m-1):
            kmj = (k-m+j) % mmax
            self.sy[k1, kmj] = self.s[k1].s_vdot(self.y[kmj])

        for i in range(m):
            kmi = (k-m+i) % mmax
            for j in range(m):
                kmj = (k-m+j) % mmax
                result[i, j] = self.ss[kmi, kmj]
                result[i, m+j] = result[m+j, i] = self.sy[kmi, kmj]
                result[m+i, m+j] = self.yy[kmi, kmj]

            sgrad_i = self.s[kmi].s_vdot(self.last_gradient)
            result[2*m, i] = result[i, 2*m] = sgrad_i

            ygrad_i = self.y[kmi].s_vdot(self.last_gradient)
            result[2*m, m+i] = result[m+i, 2*m] = ygrad_i

        result[2*m, 2*m] = self.last_gradient.norm()
        return result

    @property
    def delta(self):
        """Calculates the new scalar coefficients (deltas).

        Returns
        -------
        List
            List of the new scalar coefficients (deltas).
        """
        m = self.history_length
        b_dot_b = self.b_dot_b

        delta = np.zeros(2*m+1, dtype=np.float64)
        delta[2*m] = -1

        alpha = np.empty(m, dtype=np.float64)

        for j in range(m-1, -1, -1):
            delta_b_b = sum([delta[l] * b_dot_b[l, j] for l in range(2*m+1)])
            alpha[j] = delta_b_b/b_dot_b[j, m+j]
            delta[m+j] -= alpha[j]

        for i in range(2*m+1):
            delta[i] *= b_dot_b[m-1, 2*m-1]/b_dot_b[2*m-1, 2*m-1]

        for j in range(m):
            delta_b_b = sum([delta[l]*b_dot_b[m+j, l] for l in range(2*m+1)])
            beta = delta_b_b/b_dot_b[j, m+j]
            delta[j] += (alpha[j] - beta)

        return delta

    def add_new_point(self, x, gradient):
        """Updates the s list and y list.

        Calculates the new position and gradient differences and enters them
        into the respective list.
        """
        mmax = self.max_history_length
        self.s[self.k % mmax] = x - self.last_x
        self.y[self.k % mmax] = gradient - self.last_gradient

        self.last_x = x
        self.last_gradient = gradient

        self.k += 1



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2019 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import numpy as np

from .structured_domain import StructuredDomain


class DOFSpace(StructuredDomain):
    """Generic degree-of-freedom space. It is defined as the domain of some
    DOFDistributor.
    Its entries represent the underlying degrees of freedom of some other
    space, according to the dofdex.

    Parameters
    ----------
    dof_weights: 1-D numpy array
        A numpy array containing the multiplicity of each individual degree of
        freedom.
    """

    _needed_for_hash = ["_dvol"]

    def __init__(self, dof_weights):
        self._dvol = tuple(dof_weights)

    @property
    def harmonic(self):
        return False

    @property
    def shape(self):
        return (len(self._dvol),)

    @property
    def size(self):
        return len(self._dvol)

    @property
    def scalar_dvol(self):
        return None

    @property
    def dvol(self):
        return np.array(self._dvol)

    def __repr__(self):
        return f'DOFSpace(dof_weights={self._dvol})'



#!/usr/bin/env python3

# SPDX-License-Identifier: GPL-2.0+ OR BSD-2-Clause
# Authors: Gordian Edenhofer, Philipp Frank

import operator
from dataclasses import dataclass
from functools import reduce
from typing import Callable, Iterable, Optional

import jax.numpy as jnp
import numpy as np
import numpy.typing as npt
from jax import ShapeDtypeStruct, eval_shape
from jax.experimental import checkify
from jax.lax import select


@dataclass()
class GridAtLevel:
    shape: npt.NDArray[np.int_]
    splits: Optional[npt.NDArray[np.int_]]
    parent_splits: Optional[npt.NDArray[np.int_]]

    def __init__(self, shape, splits=None, parent_splits=None):
        self.shape = np.atleast_1d(shape)
        if splits is not None:
            splits = np.broadcast_to(splits, (self.ndim,))
        if parent_splits is not None:
            parent_splits = np.broadcast_to(parent_splits, (self.ndim,))
        self.splits = splits
        self.parent_splits = parent_splits

    def _parse_index(self, index):
        index = jnp.asarray(index)
        if index.shape[0] != self.shape.size:
            ve = f"index {index} is of invalid length {index.shape[0]} for shape {self.shape}"
            raise IndexError(ve)
        # Follow jax-array style out of bounds handling
        shp_bc = self.shape[(slice(None),) + (np.newaxis,) * (index.ndim - 1)]
        index = select(jnp.abs(index) < shp_bc, index, jnp.sign(index) * (shp_bc - 1))
        return index % shp_bc

    @property
    def size(self):
        return reduce(operator.mul, self.shape, 1)

    @property
    def ndim(self):
        return len(self.shape)

    @property
    def raw_grids(self):
        return (self,)

    def refined_indices(self):
        if self.splits is None:
            raise IndexError("this level has no children")
        # TODO non-dense grid
        return np.mgrid[tuple(slice(0, sh) for sh in self.shape)]

    def resort(self, batched_ar, /):
        if batched_ar.ndim != 2 * self.ndim:
            raise ValueError
        if batched_ar.shape[1::2] != tuple(self.parent_splits):
            raise ValueError
        shp = batched_ar.shape
        return batched_ar.reshape(tuple(a * b for a, b in zip(shp[::2], shp[1::2])))

    def children(self, index):
        if self.splits is None:
            raise IndexError("this level has no children")
        index = self._parse_index(index)
        dtp = np.result_type(index)
        f = self.splits[(slice(None),) + (np.newaxis,) * (index.ndim - 1)]
        c = np.mgrid[tuple(slice(sz) for sz in self.splits)].astype(dtp)
        c_bc = (
            (slice(None),)
            + (np.newaxis,) * (index.ndim - 1)
            + (slice(None),) * self.ndim
        )
        ids = index * f
        return ids[(slice(None),) * ids.ndim + (np.newaxis,) * self.ndim] + c[c_bc]

    def neighborhood(self, index, window_size: Iterable[int]):
        index = self._parse_index(index)
        dtp = np.result_type(index)
        window_size = np.asarray(window_size)
        assert window_size.size == self.ndim
        c = np.mgrid[tuple(slice(sz) for sz in window_size)]
        c -= (window_size // 2)[(slice(None),) + (np.newaxis,) * self.ndim]
        c_bc = (
            (slice(None),)
            + (np.newaxis,) * (index.ndim - 1)
            + (slice(None),) * self.ndim
        )
        m_bc = (slice(None),) + (np.newaxis,) * (index.ndim - 1 + self.ndim)
        id_bc = (slice(None),) * index.ndim + (np.newaxis,) * self.ndim
        res = (index[id_bc] + c[c_bc]) % self.shape[m_bc]
        return res.astype(dtp)

    def parent(self, index):
        if self.parent_splits is None:
            raise IndexError("you are alone in this world")
        index = self._parse_index(index)
        bc = (slice(None),) + (np.newaxis,) * (index.ndim - 1)
        return index // self.parent_splits[bc]

    def index2coord(self, index):
        slc = (slice(None),) + (np.newaxis,) * (index.ndim - 1)
        return (index + 0.5) / self.shape[slc]

    def coord2index(self, coord, dtype=np.uint64):
        slc = (slice(None),) + (np.newaxis,) * (coord.ndim - 1)
        index = (coord * self.shape[slc] - 0.5)
        if np.issubdtype(dtype, np.integer):
            return np.rint(index).astype(dtype)
        else:
            raise ValueError(f"non-integer index dtype: {dtype}")

    def index2volume(self, index):
        return np.array(1.0 / self.size)[(np.newaxis,) * index.ndim]


@dataclass()
class Grid:
    """Dense grid with periodic boundary conditions."""

    shape0: npt.NDArray[np.int_]
    splits: tuple[npt.NDArray[np.int_]]
    atLevel: Callable

    def __init__(self, *, shape0, splits, atLevel=GridAtLevel):
        self.shape0 = np.atleast_1d(shape0)
        splits = (splits,) if isinstance(splits, int) else splits
        self.splits = tuple(np.broadcast_to(s, self.shape0.shape) for s in splits)
        self.atLevel = atLevel

    @property
    def depth(self):
        return len(self.splits)

    def _parse_level(self, level):
        if np.abs(level) > self.depth:
            raise IndexError(f"{self.__class__.__name__} does not have level {level}")
        return level % (self.depth + 1)

    def amend(self, splits):
        splits = (splits,) if isinstance(splits, int) else splits
        splits = tuple(np.broadcast_to(s, self.shape0.shape) for s in splits)
        return self.__class__(
            shape0=self.shape0, splits=self.splits + splits, atLevel=self.atLevel
        )

    def at(self, level: int):
        level = self._parse_level(level)
        fct = 1
        if level > 0:
            fct = np.array(
                [reduce(operator.mul, si) for si in zip(*self.splits[:level])]
            )
        return self.atLevel(
            shape=self.shape0 * fct,
            splits=self.splits[level] if level < self.depth else None,
            parent_splits=self.splits[level - 1] if level >= 1 else None,
        )


@dataclass()
class OpenGridAtLevel(GridAtLevel):
    padding: Optional[npt.NDArray[np.int_]]
    parent_padding: Optional[npt.NDArray[np.int_]]
    shifts: Optional[npt.NDArray[np.int_]]

    def __init__(
        self,
        shape,
        splits=None,
        parent_splits=None,
        *,
        padding=None,
        parent_padding=None,
        shifts=None,
        level=None,
        all_splits=None,
    ):
        super().__init__(shape=shape, splits=splits, parent_splits=parent_splits)
        if padding is not None:
            padding = np.broadcast_to(padding, (self.ndim,))
        if parent_padding is not None:
            parent_padding = np.broadcast_to(parent_padding, (self.ndim,))
        if shifts is not None:
            shifts = np.broadcast_to(shifts, (self.ndim,))
        self.padding = padding
        self.parent_padding = parent_padding
        self.shifts = shifts
        del level, all_splits  # not used here but used by child classes

    def refined_indices(self):
        if self.splits is None:
            raise IndexError("ths level has no children")
        # TODO non-dense grid
        return np.mgrid[
            tuple(slice(pp, sh - pp) for sh, pp in zip(self.shape, self.padding))
        ]

    def children(self, index):
        if (self.splits is None) or (self.padding is None):
            raise IndexError("this level has no children")
        lo = self.padding[(slice(None),) + (np.newaxis,) * (index.ndim - 1)]
        hi = self.shape[(slice(None),) + (np.newaxis,) * (index.ndim - 1)] - lo
        # TODO add option for fully jax transformable error handling
        # Follow jax-array inspired out of bounds handling
        return super().children(index.clip(lo, hi - 1) - lo)

    def neighborhood(self, index, window_size: Iterable[int]):
        # TODO add option for fully jax transformable error handling
        # Follow jax-array inspired out of bounds handling
        shp_bc = self.shape[
            (slice(None),) + (np.newaxis,) * (index.ndim - 1 + self.ndim)
        ]
        return super().neighborhood(index, window_size).clip(0, (shp_bc - 1))

    def parent(self, index):
        if self.parent_splits is None:
            raise IndexError("you are alone in this world")
        index = self._parse_index(index)
        bc = (slice(None),) + (np.newaxis,) * (index.ndim - 1)
        return (index // self.parent_splits[bc]) + self.parent_padding[bc]

    def index2coord(self, index):
        slc = (slice(None),) + (np.newaxis,) * (index.ndim - 1)
        shp = self.shape + 2 * self.shifts
        index = index + self.shifts[slc]
        return (index + 0.5) / shp[slc]

    def coord2index(self, coord, dtype=np.uint64):
        slc = (slice(None),) + (np.newaxis,) * (coord.ndim - 1)
        shp = self.shape + 2 * self.shifts
        index = coord * shp[slc] - self.shifts[slc] - 0.5
        if np.issubdtype(dtype, np.integer):
            return np.rint(index).astype(dtype)
        else:
            raise ValueError(f"non-integer index dtype: {dtype}")

    def index2volume(self, index):
        sz = np.prod(self.shape + 2 * self.shifts)
        return np.array(1.0 / sz)[(np.newaxis,) * index.ndim]


@dataclass()
class OpenGrid(Grid):
    """Dense grid with open boundary conditions.

    At each level the grid has all required indices for a full
    refinement/convolution even for those that are not split. However, indices
    used for padding don't have children. The coordinates for the
    refinement/convolution span the full space including any padding on all
    previous layers.
    """

    def __init__(self, *, shape0, splits, padding, atLevel=OpenGridAtLevel):
        super().__init__(shape0=shape0, splits=splits, atLevel=atLevel)
        padding = (padding,) if isinstance(padding, int) else padding
        self.padding = tuple(np.broadcast_to(p, self.shape0.shape) for p in padding)
        if len(self.padding) != len(self.splits):
            msg = f"padding ({padding!r}) and splits ({splits!r}) not of equal length"
            raise ValueError(msg)
        # Validate that the shape is always valid (greater than zero)
        shp = self.shape0
        for si, pd in zip(self.splits, self.padding):
            shp = si * (shp - 2 * pd)
            assert np.all(shp > 0)

    def amend(self, splits, padding):
        splits = (splits,) if isinstance(splits, int) else splits
        splits = tuple(np.broadcast_to(s, self.shape0.shape) for s in splits)
        padding = (padding,) if isinstance(padding, int) else padding
        padding = tuple(np.broadcast_to(p, self.shape0.shape) for p in padding)
        return self.__class__(
            shape0=self.shape0,
            splits=self.splits + splits,
            padding=self.padding + padding,
            atLevel=self.atLevel,
        )

    def at(self, level: int):
        level = self._parse_level(level)
        shp = self.shape0
        shifts = np.zeros_like(shp)
        for si, pd in zip(self.splits[:level], self.padding[:level]):
            shp = si * (shp - 2 * pd)
            shifts = si * (shifts + pd)
        return self.atLevel(
            shape=shp,
            splits=self.splits[level] if level < self.depth else None,
            parent_splits=self.splits[level - 1] if level >= 1 else None,
            padding=self.padding[level] if level < self.depth else None,
            parent_padding=self.padding[level - 1] if level >= 1 else None,
            shifts=shifts,
            level=level,
            all_splits=self.splits[: level + 1],
        )


@dataclass()
class MGridAtLevel(GridAtLevel):
    """Multi-dimensional meshgrid product of multiple grids."""

    grids: tuple[GridAtLevel]

    def __init__(self, *grids):
        self.grids = tuple(grids)

    @property
    def shape(self):
        return np.concatenate(tuple(g.shape for g in self.grids))

    @property
    def splits(self):
        if self.grids[0].splits is None:
            if any(gg.splits is not None for gg in self.grids[1:]):
                raise ValueError(f"inconsistent `None` splits in grids {self.grids}")
            return None
        return np.concatenate(tuple(g.splits for g in self.grids))

    @property
    def parent_splits(self):
        if self.grids[0].parent_splits is None:
            if any(gg.parent_splits is not None for gg in self.grids[1:]):
                msg = f"inconsistent `None` parent_splits in grids {self.grids}"
                raise ValueError(msg)
            return None
        return np.concatenate(tuple(g.parent_splits for g in self.grids))

    @property
    def ngrids(self):
        return len(self.grids)

    @property
    def raw_grids(self):
        return reduce(operator.add, (g.raw_grids for g in self.grids))

    def refined_indices(self):
        mgrids = tuple(gg.refined_indices() for gg in self.grids)
        res = mgrids[0]
        for mg in mgrids[1:]:
            slf = (slice(None),) * res.ndim + (jnp.newaxis,) * (mg.ndim - 1)
            slb = (
                (slice(None),)
                + (jnp.newaxis,) * (res.ndim - 1)
                + (slice(None),) * (mg.ndim - 1)
            )
            shb = res.shape[1:] + mg.shape[1:]
            res = jnp.broadcast_to(res[slf], (res.shape[0],) + shb)
            mg = jnp.broadcast_to(mg[slb], (mg.shape[0],) + shb)
            res = jnp.concatenate((res, mg), axis=0)
        return res

    def resort(self, batched_ar, /):
        if any(isinstance(g, FlatGrid) for g in self.grids):
            raise NotImplementedError()  # TODO generalize using grids resort
        return super().resort(batched_ar)

    def children(self, index) -> npt.NDArray:
        ndims_off = tuple(np.cumsum(tuple(g.ndim for g in self.grids)))
        ndims_sum = ndims_off[-1]
        islice = tuple(slice(l, r) for l, r in zip((0,) + ndims_off[:-1], ndims_off))
        children = tuple(g.children(index[i]) for i, g in zip(islice, self.grids))
        # Make initial entry broadcast-able to the full final shape
        out = children[0][
            (slice(None),) * children[0].ndim
            + (np.newaxis,) * (ndims_sum - self.grids[0].ndim)
        ]
        # Successively concatenate all broadcasted children
        for c, i in zip(children[1:], islice[1:]):
            c = c[
                (slice(None),) * index.ndim
                + (np.newaxis,) * i.start
                + (slice(None),) * (i.stop - i.start)
                + (np.newaxis,) * (ndims_sum - i.stop)
            ]
            assert c.shape[0] == (i.stop - i.start)
            bshp = np.broadcast_shapes(out.shape[1:], c.shape[1:])
            c = jnp.broadcast_to(c, c.shape[:1] + bshp)
            out = jnp.broadcast_to(out, out.shape[:1] + bshp)
            out = jnp.concatenate((out, c), axis=0)
        return out

    def neighborhood(self, index, window_size: tuple[int]):
        ndims_off = tuple(np.cumsum(tuple(g.ndim for g in self.grids)))
        ndims_sum = ndims_off[-1]
        islice = tuple(slice(l, r) for l, r in zip((0,) + ndims_off[:-1], ndims_off))
        window_size = (
            (window_size,) * self.ndim if isinstance(window_size, int) else window_size
        )
        assert len(window_size) == self.ndim
        neighborhood = []
        for i, g in zip(islice, self.grids):
            n = g.neighborhood(index[i], window_size[i])
            neighborhood.append(n)

        # Make initial entry broadcast-able to the full final shape
        out = neighborhood[0][
            (slice(None),) * neighborhood[0].ndim
            + (np.newaxis,) * (ndims_sum - self.grids[0].ndim)
        ]
        # Successively concatenate all broadcasted neighbors
        for n, i in zip(neighborhood[1:], islice[1:]):
            n = n[
                (slice(None),) * index.ndim
                + (np.newaxis,) * i.start
                + (slice(None),) * (i.stop - i.start)
                + (np.newaxis,) * (ndims_sum - i.stop)
            ]
            assert n.shape[0] == (i.stop - i.start)
            bshp = np.broadcast_shapes(out.shape[1:], n.shape[1:])
            n = jnp.broadcast_to(n, n.shape[:1] + bshp)
            out = jnp.broadcast_to(out, out.shape[:1] + bshp)
            out = jnp.concatenate((out, n), axis=0)
        return out

    def parent(self, index):
        ndims_off = tuple(np.cumsum(tuple(g.ndim for g in self.grids)))
        islice = tuple(slice(l, r) for l, r in zip((0,) + ndims_off[:-1], ndims_off))
        parent = tuple(g.parent(index[i]) for i, g in zip(islice, self.grids))
        return jnp.concatenate(parent, axis=0)

    def index2coord(self, index, **kwargs):
        ndims_off = tuple(np.cumsum(tuple(g.ndim for g in self.grids)))
        islice = tuple(slice(l, r) for l, r in zip((0,) + ndims_off[:-1], ndims_off))
        coord = tuple(g.index2coord(index[i]) for i, g in zip(islice, self.grids))
        return jnp.concatenate(coord, axis=0)

    def coord2index(self, coord, **kwargs):
        cdims = tuple(
            eval_shape(gg.index2coord, ShapeDtypeStruct((gg.ndim,), jnp.int_)).shape[0]
            for gg in self.grids
        )
        ndims_off = tuple(np.cumsum(cdims))
        islice = tuple(slice(l, r) for l, r in zip((0,) + ndims_off[:-1], ndims_off))
        index = tuple(g.coord2index(coord[i]) for i, g in zip(islice, self.grids))
        return jnp.concatenate(index, axis=0)

    def index2volume(self, index, **kwargs):
        ndims_off = tuple(np.cumsum(tuple(g.ndim for g in self.grids)))
        islice = tuple(slice(l, r) for l, r in zip((0,) + ndims_off[:-1], ndims_off))
        volume = tuple(g.index2volume(index[i]) for i, g in zip(islice, self.grids))
        return reduce(operator.mul, volume)


@dataclass()
class MGrid(Grid):
    """Multi-dimensional meshgrid product of multiple grids."""

    grids: tuple[Grid]

    def __init__(self, *grids, atLevel=MGridAtLevel):
        self.grids = tuple(grids)
        for i, g in enumerate(grids):
            if not isinstance(g, Grid):
                raise TypeError(
                    f"Grid {g.__class__.__name__} at index {i} of invalid type"
                )
            if g.depth != grids[0].depth:
                raise ValueError(
                    f"Grid {g.__class__.__name__} at index {i} not of same depth"
                )
        shape0 = np.concatenate(tuple(g.shape0 for g in self.grids))
        splits = tuple(
            np.concatenate(tuple(g.splits[lvl] for g in self.grids))
            for lvl in range(self.grids[0].depth)
        )
        super().__init__(shape0=shape0, splits=splits, atLevel=atLevel)

    @property
    def depth(self):
        return self.grids[0].depth

    @property
    def ngrids(self):
        return len(self.grids)

    def amend(self, splits):
        splits = (splits,) if isinstance(splits, int) else splits
        splits = tuple(np.broadcast_to(s, self.shape0.shape) for s in splits)
        # Create slices for indexing individal grid regions in split
        ndims_off = tuple(np.cumsum(tuple(g.ndim for g in self.grids)))
        islice = tuple(slice(l, r) for l, r in zip((0,) + ndims_off[:-1], ndims_off))
        # Separate shared splits into splits for the individual grids
        spg = tuple(zip(*(tuple(spl[i] for i in islice) for spl in splits)))
        grids = tuple(g.amend(s) for g, s in zip(self.grids, spg))
        return self.__class__(*grids, atLevel=self.atLevel)

    def at(self, level: int):
        level = self._parse_level(level)
        return self.atLevel(*tuple(g.at(level) for g in self.grids))


@dataclass()
class FlatGridAtLevel(GridAtLevel):
    """Same as :class:`GridAtLevel` but with a single global integer index for each voxel."""

    grid_at_level: GridAtLevel
    all_shapes: npt.NDArray[np.int_]
    all_splits: tuple[npt.NDArray[np.int_]]
    ordering: str

    def __init__(self, grid_at_level, *, all_shapes, all_splits, ordering="serial"):
        if not isinstance(grid_at_level, GridAtLevel):
            raise TypeError(f"Grid {grid_at_level.__name__} of invalid type")
        self.grid_at_level = grid_at_level
        ordering = str(ordering).lower()
        if ordering not in ("serial", "nest"):
            raise ValueError(f"invalid flat index ordering scheme {ordering}")
        self.ordering = ordering
        self.all_shapes = tuple(np.atleast_1d(sh) for sh in all_shapes)
        self.all_splits = tuple(
            np.broadcast_to(sp, shp.shape)
            for sp, shp in zip(all_splits, self.all_shapes)
        )
        super().__init__(
            shape=(reduce(operator.mul, grid_at_level.shape, 1),),
            splits=None,
            parent_splits=None,
        )

    @property
    def raw_grids(self):
        return self.grid_at_level.raw_grids

    def _weights_serial(self, levelshift):
        if levelshift not in (-1, 0, 1):
            raise ValueError(f"invalid shift in level {levelshift!r}")
        shape = self.all_shapes[-2 + levelshift]
        return np.cumprod(np.append(shape[1:], 1)[::-1])[::-1]

    def _weights_nest(self, levelshift):
        raise NotImplementedError

    def index2flatindex(self, index, levelshift=0):
        # TODO vectorize better
        if self.ordering == "serial":
            wgt = self._weights_serial(levelshift)
            wgt = wgt[(slice(None),) + (np.newaxis,) * (index.ndim - 1)]
            return (wgt * index).sum(axis=0).astype(index.dtype)[jnp.newaxis, ...]
        if self.ordering == "nest":
            raise NotImplementedError
        raise RuntimeError

    def flatindex2index(self, index, levelshift=0):
        # TODO vectorize better
        dtp = index.dtype
        if self.ordering == "serial":
            wgt = self._weights_serial(levelshift)
            tm = jnp.copy(index[0])
            index = np.zeros(wgt.shape + index.shape[1:], dtype=dtp)
            index = []
            for w in wgt:
                tmfl = tm // w
                tm -= w * tmfl
                index.append(tmfl)
            return jnp.stack(index, axis=0).astype(dtp)
        if self.ordering == "nest":
            raise NotImplementedError
        raise RuntimeError

    def refined_indices(self):
        ids = self.grid_at_level.refined_indices()
        return self.index2flatindex(ids).reshape((1, -1))

    def resort(self, batched_ar, /):
        parent_splits = self.all_splits[-3]
        shape = self.all_shapes[-2]
        if batched_ar.ndim != 2:
            raise ValueError
        if batched_ar.shape[1] != np.prod(parent_splits):
            raise ValueError

        if self.ordering == "serial":
            shp = tuple(shape // parent_splits) + tuple(parent_splits)
            batched_ar = batched_ar.reshape(shp)
            ndim = shape.size
            idx = np.arange(ndim)
            axes = reduce(operator.add, ((a, b) for a, b in zip(idx, ndim + idx)))
            return jnp.transpose(batched_ar, axes).ravel()
        if self.ordering == "nest":
            raise NotImplementedError
        raise RuntimeError

    def children(self, index) -> npt.NDArray:
        index = self._parse_index(index)
        index = self.flatindex2index(index)
        children = self.grid_at_level.children(index).reshape(index.shape + (-1,))
        return self.index2flatindex(children, +1)

    def neighborhood(self, index, window_size: Iterable[int]):
        index = self._parse_index(index)
        index = self.flatindex2index(index)
        window = self.grid_at_level.neighborhood(index, window_size=window_size)
        return self.index2flatindex(window.reshape(index.shape + (-1,)))

    def parent(self, index):
        index = self._parse_index(index)
        index = self.flatindex2index(index)
        window = self.grid_at_level.parent(index)
        return self.index2flatindex(window, -1)

    def index2coord(self, index, **kwargs):
        index = self._parse_index(index)
        index = self.flatindex2index(index)
        return self.grid_at_level.index2coord(index, **kwargs)

    def coord2index(self, coord, **kwargs):
        index = self.grid_at_level.coord2index(coord, **kwargs)
        return self.index2flatindex(index)

    def index2volume(self, index, **kwargs):
        index = self._parse_index(index)
        index = self.flatindex2index(index)
        return self.grid_at_level.index2volume(index, **kwargs)


@dataclass()
class FlatGrid(Grid):
    """Same as :class:`Grid` but with a single global integer index for each voxel."""

    grid: Grid
    ordering: str

    def __init__(self, grid, *, ordering="serial", atLevel=FlatGridAtLevel):
        if not isinstance(grid, Grid):
            raise TypeError(f"Grid {grid.__name__} of invalid type")
        self.grid = grid
        ordering = str(ordering).lower()
        if ordering not in ("serial", "nest"):
            raise ValueError(f"invalid flat index ordering scheme {ordering}")
        self.ordering = ordering
        shape0 = np.array([np.prod(grid.shape0)])
        splits = tuple(np.array([np.prod(spl)]) for spl in grid.splits)
        super().__init__(shape0=shape0, splits=splits, atLevel=atLevel)

    def amend(self, splits):
        grid = self.grid.amend(splits)
        return self.__class__(grid=grid, ordering=self.ordering)

    def at(self, level: int):
        level = self._parse_level(level)
        shapes = []
        splits = []
        for lvl in range(level + 2):
            if lvl <= self.depth:
                atlvl = self.grid.at(lvl)
                shapes.append(atlvl.shape)
                splits.append(atlvl.splits)
            else:
                shapes.append(None)
                splits.append(None)
        return self.atLevel(
            self.grid.at(level),
            all_shapes=shapes,
            all_splits=splits,
            ordering=self.ordering,
        )


@dataclass()
class SparseGridAtLevel(FlatGridAtLevel):
    mapping: npt.NDArray[np.int_]
    parent_mapping: Optional[npt.NDArray[np.int_]] = None
    children_mapping: Optional[npt.NDArray[np.int_]] = None

    def __init__(
        self,
        grid_at_level,
        all_shapes,
        all_splits,
        mapping,
        ordering="nest",
        parent_mapping=None,
        children_mapping=None,
    ):
        if not isinstance(grid_at_level, GridAtLevel):
            raise TypeError(f"Grid {grid_at_level.__name__} of invalid type")
        self.mapping = mapping
        self.parent_mapping = parent_mapping
        self.children_mapping = children_mapping
        super().__init__(
            grid_at_level=grid_at_level,
            all_shapes=all_shapes,
            all_splits=all_splits,
            ordering=ordering,
        )
        # Overrides shape to utilize base functions
        self.shape = np.array([self.mapping.size])

    def _mapping(self, levelshift):
        if levelshift == -1:
            mapping = self.parent_mapping
        elif levelshift == 0:
            mapping = self.mapping
        elif levelshift == 1:
            mapping = self.children_mapping
        else:
            raise ValueError(f"invalid shift in level: {levelshift}")
        return mapping

    def arrayindex2flatindex(self, index, levelshift=0):
        index = self._parse_index(index)
        return self._mapping(levelshift)[index]

    def flatindex2arrayindex(self, index, levelshift=0):
        mapping = self._mapping(levelshift)
        arrayid = np.searchsorted(mapping, index)
        #  TODO Benchmark searchsorted on stack instead of second one with `right`
        valid = np.searchsorted(mapping, index, side="right") == arrayid + 1

        checkify.check(
            jnp.all(valid),
            f"flat index {{ids}} not on child grid of {self.__class__.__name__}",
            ids=arrayid[~valid],
        )
        return arrayid

    def refined_indices(self):
        raise NotImplementedError  # TODO

    def children(self, index) -> npt.NDArray:
        index = self.arrayindex2flatindex(index)
        index = self.flatindex2index(index)
        children = self.grid_at_level.children(index)
        children = self.index2flatindex(children, +1)
        return self.flatindex2arrayindex(children, +1)

    def neighborhood(self, index, window_size: Iterable[int]):
        window = self.arrayindex2flatindex(index)
        window = self.flatindex2index(window)
        window = self.grid_at_level.neighborhood(index, window_size=window_size)
        window = self.index2flatindex(window)
        return self.flatindex2arrayindex(window)

    def parent(self, index):
        index = self.arrayindex2flatindex(index)
        index = self.flatindex2index(index)
        parent = self.grid_at_level.parent(index)
        parent = self.index2flatindex(parent, -1)
        return self.flatindex2arrayindex(parent, -1)

    def index2coord(self, index, **kwargs):
        index = self.arrayindex2flatindex(index)
        index = self.flatindex2index(index)
        return self.grid_at_level.index2coord(index, **kwargs)

    def coord2index(self, coord, **kwargs):
        index = self.grid_at_level.coord2index(coord, **kwargs)
        index = self.index2flatindex(index)
        return self.flatindex2arrayindex(index)

    def index2volume(self, index, **kwargs):
        index = self.arrayindex2flatindex(index)
        index = self.flatindex2index(index)
        return self.grid_at_level.index2volume(index, **kwargs)

    def to_flat_grid(self):
        # FIXME @ph-frank incompatible with the current `FlatGridAtLevel` impl
        return FlatGridAtLevel(
            self.grid_at_level,
            self.shape0,
            self.all_parent_splits,
            ordering=self.ordering,
        )


@dataclass()
class SparseGrid(FlatGrid):
    """Same as :class:`FlatGrid` but keeping track of the indices that are actually
    being modeled. This class is especially convenient for open boundary conditions
    but works for arbitrarily sparsely resolved grids."""

    mapping: tuple[npt.NDArray[np.int_]]

    def __init__(
        self,
        grid,
        mapping,
        ordering="nest",
        *,
        _check_mapping=True,
        atLevel=SparseGridAtLevel,
    ):
        if not isinstance(grid, Grid):
            raise TypeError(f"Grid {grid.__class__.__name__} of invalid type")
        self.grid = grid
        self.ordering = ordering

        mapping = (mapping,) if not isinstance(mapping, tuple) else mapping
        mapping = tuple(np.atleast_1d(m) for m in mapping)

        if _check_mapping:
            if len(mapping) != grid.depth + 1:
                md, gd = len(mapping), grid.depth
                nm = grid.__class__.__name__
                msg = f"Map depth {md} does not match grid {nm} depth {gd}"
                raise ValueError(msg)
            for mm in mapping:
                if mm.ndim != 1:
                    raise IndexError("Mapping must be one dimensional")
                if np.any(mm[1:] <= mm[:-1]):
                    raise IndexError("Mapping must be unique and sorted")
        self.mapping = mapping

        super().__init__(grid=grid, ordering=ordering, atLevel=atLevel)

    def amend(self, splits, mapping, **kwargs):
        grid = self.grid.amend(splits, **kwargs)
        mapping = (mapping,) if not isinstance(mapping, tuple) else mapping
        return self.__class__(grid, mapping, nest=self.nest)

    def at(self, level: int):
        level = self._parse_level(level)
        grid_at_level = self.grid.at(level)
        parent_mapping = None if level == 0 else self.mapping[level - 1]
        children_mapping = None if level == self.depth else self.mapping[level + 1]
        return self.atLevel(
            grid_at_level,
            # FIXME @ph-frank pass `all_shapes`,
            self.grid.splits[: (level + 1)] + ((None,) if level == self.depth else ()),
            self.mapping[level],
            ordering=self.ordering,
            parent_mapping=parent_mapping,
            children_mapping=children_mapping,
        )



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2020 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

from .linear_operator import LinearOperator


class EndomorphicOperator(LinearOperator):
    """Represents a :class:`LinearOperator` which is endomorphic, i.e. one
    which has identical domain and target.
    """
    @property
    def target(self):
        """DomainTuple : returns :attr:`domain`

        Returns `self.domain`, because this is also the target domain
        for endomorphic operators."""
        return self._domain

    def draw_sample(self, from_inverse=False, device_id=-1):
        """Generates a sample from a Gaussian distribution with zero mean and
        covariance given by the operator.

        May or may not be implemented. Only optional.

        Parameters
        ----------
        from_inverse : bool (default : False)
            if True, the sample is drawn from the inverse of the operator

        device_id : int
            Device id on which the Field shall be located. Default: -1

        Returns
        -------
        :class:`nifty.cl.field.Field` or :class:`nifty.cl.multi_field.MultiField`
            A sample from the Gaussian of given covariance.
        """
        raise NotImplementedError

    @property
    def sampling_dtype(self):
        """Sampling dtype if operator is used as covariance operator."""
        if hasattr(self, "_dtype"):
            return self._dtype
        return None

    def get_sqrt(self):
        """Return operator op which obeys `self == op.adjoint @ op`.

        Note that this function is only implemented for operators with real
        spectrum.

        Returns
        -------
        EndomorphicOperator
            Operator which is the square root of `self`

        """
        raise NotImplementedError

    def _dom(self, mode):
        return self._domain

    def _tgt(self, mode):
        return self._domain

    def _check_input(self, x, mode):
        self._check_mode(mode)
        if self.domain != x.domain:
            raise ValueError("The operator's and field's domains don't match.")



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2020 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

from ..utilities import check_object_identity
from .operator import Operator


class LinearOperator(Operator):
    """NIFTY base class for linear operators.

    The base NIFTY operator class is an abstract class from which
    other specific operator subclasses are derived.

    Attributes
    ----------
    TIMES : int
        Symbolic constant representing normal operator application
    ADJOINT_TIMES : int
        Symbolic constant representing adjoint operator application
    INVERSE_TIMES : int
        Symbolic constant representing inverse operator application
    ADJOINT_INVERSE_TIMES : int
        Symbolic constant representing adjoint inverse operator application
    INVERSE_ADJOINT_TIMES : int
        same as ADJOINT_INVERSE_TIMES

    Notes
    -----
    The symbolic constants for the operation modes can be combined by the
    "bitwise-or" operator "|", for expressing the capability of the operator
    by means of a single integer number.
    """

    # Field Operator Modes
    TIMES = 1
    ADJOINT_TIMES = 2
    INVERSE_TIMES = 4
    ADJOINT_INVERSE_TIMES = 8
    INVERSE_ADJOINT_TIMES = 8

    # Operator Transform Flags
    ADJOINT_BIT = 1
    INVERSE_BIT = 2

    _ilog = (-1, 0, 1, -1, 2, -1, -1, -1, 3)
    _validMode = (False, True, True, False, True, False, False, False, True)
    _modeTable = ((1, 2, 4, 8),
                  (2, 1, 8, 4),
                  (4, 8, 1, 2),
                  (8, 4, 2, 1))
    _capTable = ((0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15),
                 (0, 2, 1, 3, 8, 10, 9, 11, 4, 6, 5, 7, 12, 14, 13, 15),
                 (0, 4, 8, 12, 1, 5, 9, 13, 2, 6, 10, 14, 3, 7, 11, 15),
                 (0, 8, 4, 12, 2, 10, 6, 14, 1, 9, 5, 13, 3, 11, 7, 15))
    _addInverse = (0, 5, 10, 15, 5, 5, 15, 15, 10, 15, 10, 15, 15, 15, 15, 15)
    _backwards = 6
    _all_ops = 15

    def _dom(self, mode):
        return self.domain if (mode & 9) else self.target

    def _tgt(self, mode):
        return self.domain if (mode & 6) else self.target

    def _flip_modes(self, trafo):
        from .operator_adapter import OperatorAdapter
        return self if trafo == 0 else OperatorAdapter(self, trafo)

    @property
    def inverse(self):
        """LinearOperator : the inverse of `self`

        Returns a LinearOperator object which behaves as if it were
        the inverse of this operator."""
        return self._flip_modes(self.INVERSE_BIT)

    @property
    def adjoint(self):
        """LinearOperator : the adjoint of `self`

        Returns a LinearOperator object which behaves as if it were
        the adjoint of this operator."""
        return self._flip_modes(self.ADJOINT_BIT)

    def __matmul__(self, other):
        if other.isIdentity():
            return self
        if isinstance(other, LinearOperator):
            from .chain_operator import ChainOperator
            return ChainOperator.make([self, other])
        return Operator.__matmul__(self, other)

    def __rmatmul__(self, other):
        if other.isIdentity():
            return self
        if isinstance(other, LinearOperator):
            from .chain_operator import ChainOperator
            return ChainOperator.make([other, self])
        return Operator.__rmatmul__(self, other)

    def _myadd(self, other, oneg):
        from .sum_operator import SumOperator
        return SumOperator.make((self, other), (False, oneg))

    def __add__(self, other):
        if isinstance(other, LinearOperator):
            return self._myadd(other, False)
        return Operator.__add__(self, other)

    def __radd__(self, other):
        return self.__add__(other)

    def __sub__(self, other):
        if isinstance(other, LinearOperator):
            return self._myadd(other, True)
        return Operator.__sub__(self, other)

    def __rsub__(self, other):
        if isinstance(other, LinearOperator):
            return other._myadd(self, True)
        return Operator.__rsub__(self, other)

    @property
    def capability(self):
        """int : the supported operation modes

        Returns the supported subset of :attr:`TIMES`, :attr:`ADJOINT_TIMES`,
        :attr:`INVERSE_TIMES`, and :attr:`ADJOINT_INVERSE_TIMES`,
        joined together by the "|" operator.
        """
        return self._capability

    def force(self, x):
        return self.apply(x.extract(self.domain), self.TIMES)

    def apply(self, x, mode):
        """Applies the Operator to a given `x`, in a specified `mode`.

        Parameters
        ----------
        x : :class:`nifty.cl.field.Field`
            The input Field, defined on the Operator's domain or target,
            depending on mode.

        mode : int
            - :attr:`TIMES`: normal application
            - :attr:`ADJOINT_TIMES`: adjoint application
            - :attr:`INVERSE_TIMES`: inverse application
            - :attr:`ADJOINT_INVERSE_TIMES` or
              :attr:`INVERSE_ADJOINT_TIMES`: adjoint inverse application

        Returns
        -------
        :class:`nifty.cl.field.Field`
            The processed Field defined on the Operator's target or domain,
            depending on mode.
        """
        raise NotImplementedError

    def __call__(self, x):
        """Same as :meth:`times`"""
        if self.isIdentity():
            return x
        if x.jac is not None:
            return x.new(self(x._val), self).prepend_jac(x.jac)
        if x.val is not None:
            return self.apply(x, self.TIMES)
        return self@x

    def times(self, x):
        """Applies the Operator to a given Field.

        Parameters
        ----------
        x : :class:`nifty.cl.field.Field`
            The input Field, defined on the Operator's domain.

        Returns
        -------
        :class:`nifty.cl.field.Field`
            The processed Field defined on the Operator's target domain.
        """
        return self.apply(x, self.TIMES)

    def inverse_times(self, x):
        """Applies the inverse Operator to a given Field.

        Parameters
        ----------
        x : :class:`nifty.cl.field.Field`
            The input Field, defined on the Operator's target domain

        Returns
        -------
        :class:`nifty.cl.field.Field`
            The processed Field defined on the Operator's domain.
        """
        return self.apply(x, self.INVERSE_TIMES)

    def adjoint_times(self, x):
        """Applies the adjoint-Operator to a given Field.

        Parameters
        ----------
        x : :class:`nifty.cl.field.Field`
            The input Field, defined on the Operator's target domain

        Returns
        -------
        :class:`nifty.cl.field.Field`
            The processed Field defined on the Operator's domain.
        """
        return self.apply(x, self.ADJOINT_TIMES)

    def adjoint_inverse_times(self, x):
        """Applies the adjoint-inverse Operator to a given Field.

        Parameters
        ----------
        x : :class:`nifty.cl.field.Field`
            The input Field, defined on the Operator's domain.

        Returns
        -------
        :class:`nifty.cl.field.Field`
            The processed Field defined on the Operator's target domain.

        Notes
        -----
        If the operator has an `inverse` then the inverse adjoint is identical
        to the adjoint inverse. We provide both names for convenience.
        """
        return self.apply(x, self.ADJOINT_INVERSE_TIMES)

    def inverse_adjoint_times(self, x):
        """Same as :meth:`adjoint_inverse_times`"""
        return self.apply(x, self.ADJOINT_INVERSE_TIMES)

    def _check_mode(self, mode):
        if not self._validMode[mode]:
            raise NotImplementedError("invalid operator mode specified")
        if mode & self.capability == 0:
            raise NotImplementedError(
                "requested operator mode is not supported")

    def _check_input(self, x, mode):
        self._check_mode(mode)
        check_object_identity(self._dom(mode), x.domain)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2020 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import numpy as np

from ..domain_tuple import DomainTuple
from ..field import Field
from .linear_operator import LinearOperator


class OuterProduct(LinearOperator):
    """Performs the point-wise outer product of two fields.

    Parameters
    ---------
    domain : DomainTuple, the domain of the input field
    field : :class:`nifty.cl.field.Field`
    ---------
    """
    def __init__(self, domain, field):
        self._domain = DomainTuple.make(domain)
        self._field = field
        self._target = DomainTuple.make(
            tuple(sub_d for sub_d in field.domain._dom + self._domain._dom))
        self._capability = self.TIMES | self.ADJOINT_TIMES

    def _device_preparation(self, x, mode):
        self._field = self._field.at(x.device_id)

    def apply(self, x, mode):
        self._check_input(x, mode)
        self._device_preparation(x, mode)
        if mode == self.TIMES:
            res = np.multiply.outer(self._field.val, x.val)
        else:
            axes = len(self._field.shape)
            res = np.tensordot(self._field.val, x.val, axes)
        return Field(self._tgt(mode), res)



#!/usr/bin/env python3

# Copyright(C) 2013-2021 Max-Planck-Society
# SPDX-License-Identifier: GPL-2.0+ OR BSD-2-Clause

# %% [markdown]
# # Demonstration of the Wiener filter and Gaussian process field model in NIFTy.re

# ## The Model

# %%
import jax
import matplotlib.pyplot as plt
import nifty.re as jft
from jax import numpy as jnp
from jax import random

jax.config.update("jax_enable_x64", True)

seed = 42
key = random.PRNGKey(seed)


class FixedPowerCorrelatedField(jft.Model):
    def __init__(self, shape, *, distances=None):
        self.shape = (shape,) if isinstance(shape, int) else shape
        distances = (
            tuple((1 / s for s in self.shape)) if distances is None else distances
        )
        self.grid = jft.correlated_field.make_grid(
            dims, distances=distances, harmonic_type="fourier"
        )

        super().__init__(domain=jax.ShapeDtypeStruct(shape=dims, dtype=jnp.float64))

    def amplitude_spectrum(self):
        k = self.grid.harmonic_grid.mode_lengths
        return 0.02 / (1 + k**2)

    def correlate(self, x):
        ht = jft.correlated_field.hartley
        harmonic_dvol = 1 / self.grid.total_volume

        # Will be inlined during JIT and thus it is fine to get the amplitude
        # anew in every call
        a = self.amplitude_spectrum()
        a = a[self.grid.harmonic_grid.power_distributor]

        return harmonic_dvol * ht(a * x)

    def __call__(self, x):
        return self.correlate(x)


dims = (128, 128)
signal = FixedPowerCorrelatedField(dims)

# %% [markdown]
# ## The likelihood

# %%
signal_response = signal
noise_cov = lambda x: 0.1**2 * x
noise_cov_inv = lambda x: 0.1**-2 * x

# Create synthetic data
key, subkey = random.split(key)
pos_truth = jft.random_like(subkey, signal_response.domain)
signal_response_truth = signal_response(pos_truth)
key, subkey = random.split(key)
noise_truth = (
    (noise_cov(jft.ones_like(signal_response.target))) ** 0.5
) * jft.random_like(key, signal_response.target)
data = signal_response_truth + noise_truth

lh = jft.Gaussian(data, noise_cov_inv).amend(signal_response)

# %% [markdown]
# ## The Wiener filter

# %%
delta = 1e-6
key, k_w = random.split(key)
samples, info = jft.wiener_filter_posterior(
    lh,
    key=k_w,
    n_samples=20,
    draw_linear_kwargs=dict(
        cg_name="W",
        cg_kwargs=dict(absdelta=delta * jft.size(lh.domain) / 10.0, maxiter=100),
    ),
)

# %%
post_mean, post_std = jft.mean_and_std(tuple(signal(s) for s in samples))

to_plot = [
    ("Signal", signal(pos_truth), "im"),
    ("Noise", noise_truth, "im"),
    ("Data", data, "im"),
    ("Posterior Mean", post_mean, "im"),
    ("Posterior Standard Deviation", post_std, "im"),
]
fig, axs = plt.subplots(2, 3, figsize=(16, 9))
for ax, v in zip(axs.flat, to_plot):
    title, field, tp, *labels = v
    ax.set_title(title)
    if tp == "im":
        end = tuple(n * d for n, d in zip(signal.grid.shape, signal.grid.distances))
        im = ax.imshow(field.T, cmap="inferno", extent=(0.0, end[0], 0.0, end[1]))
        plt.colorbar(im, ax=ax, orientation="horizontal")
    else:
        ax_plot = ax.loglog if tp == "loglog" else ax.plot
        x = field[0]
        for f in field[1:]:
            ax_plot(x, f, alpha=0.7)
for ax in axs.flat[len(to_plot) :]:
    ax.set_axis_off()
fig.tight_layout()
fig.savefig("wiener_filter.png", dpi=400)
plt.show()



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2024-2025 Philipp Arras
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import contextlib
from functools import reduce
from itertools import chain
from operator import mul
from warnings import warn

import numpy as np

from ..config import _config
from .ducc_dispatch import vdot as cpu_vdot
from .logger import logger
from .utilities import assert_device_available, device_available, myassert

ALLOWED_WRAPPEES = [np.ndarray]
if device_available():
    import cupy
    ALLOWED_WRAPPEES += [cupy.ndarray]
ALLOWED_WRAPPEES = tuple(ALLOWED_WRAPPEES)


def _make_scalar_if_scalar(a):
    if a.shape == tuple():
        if device_available() and isinstance(a, cupy.ndarray):
            return cupy.asnumpy(a)[()]
        else:
            return a[()]
    return a


@contextlib.contextmanager
def assert_no_device_copies():
    from ..config import update
    old = _config["fail_on_device_copy"]
    update("fail_on_device_copy", True)
    try:
        yield
    finally:
        update("fail_on_device_copy", old)


class AnyArray(np.lib.mixins.NDArrayOperatorsMixin):
    """Wrapper for numpy and cupy arrays for flexible computations on CPU and GPUs

    AnyArray supports both CPU (`numpy.ndarray`) and GPU (`cupy.ndarray`)
    arrays. It provides automatic device-aware operations and conversions. It
    implements numpy-like operations such as element-wise operations, vector dot
    products, and norms. Additionally it can also be directly plugged into numpy
    functions (like `np.fft.fftn` or `np.transpose`) - the work will be
    dispatched correctly to numpy and cupy.

    `ALLOWED_WRAPPEES` are `numpy.ndarray` and, if cupy is install installed and
    GPUs are available, also `cupy.ndarray`.

    Parameters
    ----------
    arr: ALLOWED_WRAPPEES or AnyArray
        The input array that will be wrapped into an `AnyArray` instance. It can
        be either a `numpy.ndarray` (for CPU-based computations) or a
        `cupy.ndarray` (for GPU-based computations, if `cupy` is available). If
        `arr` is already an `AnyArray`, it is returned as is without
        modification. Scalars are automatically converted into `numpy.ndarray`
        instances. The stored array retains its original memory location, and no
        implicit device transfers are performed during initialization.

    """

    # ---Constructors---
    def __new__(cls, obj=None, *args, **kwargs):
        if isinstance(obj, AnyArray):
            return obj
        return super().__new__(cls)

    def __init__(self, arr):
        if isinstance(arr, AnyArray):
            return
        if np.isscalar(arr):
            arr = np.array(arr)
        if not isinstance(arr, ALLOWED_WRAPPEES):
            raise TypeError(f"arr must be in {ALLOWED_WRAPPEES}, got {arr}")
        if isinstance(arr, np.ndarray):
            self._device_id = -1
        elif isinstance(arr, cupy.ndarray):
            self._device_id = arr.device.id
        self._val = arr
        self._shape = arr.shape
        self._strides = arr.strides
        if self._val.shape == tuple():
            assert not isinstance(self._val[()], AnyArray)
        self._writeable = True
        if (
            _config["fail_on_nontrivial_anyarray_creation_on_host"]
            and (self._device_id == -1)
            and (self.size > 1)
        ):
            raise RuntimeError(
                "forbidden AnyArray creation on host: "
                f"shape {self._shape}, device_id {self._device_id}"
            )

    @staticmethod
    def full(shape, val, device_id=-1):
        """Create an AnyArray filled with a specified scalar value.

        This method generates an `AnyArray` of the given shape, filled with the
        specified scalar value. The resulting array can be placed on a specific
        device (CPU or GPU) by specifying `device_id`.

        Parameters
        ----------
        shape : tuple of int
            The shape of the output array.
        val : scalar (int, float, complex)
            The value to fill the array with. Must be a real or complex scalar.
        device_id : int, optional (default: -1)
            The target device for the array:
            - `-1` for CPU (`numpy.ndarray`).
            - `>=0` for GPU (`cupy.ndarray` with the specified device ID).

        Returns
        -------
        AnyArray
            An `AnyArray` instance of the specified shape, filled with `val`,
            located on the specified device.

        Examples
        --------
        Create a 3x3 array filled with 5 on the CPU:

        >>> AnyArray.full((3, 3), 5)

        Create a 2x2 complex array filled with `2 + 3j` on GPU (device 0):

        >>> AnyArray.full((2, 2), 2 + 3j, device_id=0)

        """
        if not np.isscalar(val):
            raise TypeError("val must be a scalar")
        if not (np.isreal(val) or np.iscomplex(val)):
            raise TypeError("need arithmetic scalar")
        xp = np if device_id == -1 else cupy
        return AnyArray(np.broadcast_to(xp.array(val), shape))

    # ---Views, copies, rights, etc.---
    def lock(self):
        """Make the AnyArray instance read-only.

        This method sets the underlying array to be non-writable, preventing any
        modifications to its contents. For `numpy.ndarray`, it updates the
        `writeable` flag to `False`.

        Once locked, the array cannot be modified unless a writable copy is
        created.

        Note
        ----
        For `cupy.ndarray`, write protection is not yet implemented due to
        limitations in `cupy` (see https://github.com/cupy/cupy/issues/2616).
        However, we still try to track the read-only status via an internal
        attribute (`AnyArray._writeable`).

        Examples
        --------
        >>> arr = AnyArray(np.array([1, 2, 3]))
        >>> arr.lock()
        >>> arr[0] = 10  # Raises ValueError

        """
        if isinstance(self, np.ndarray):
            self._val.flags.writeable = False
        # TODO: Set writable to False for cupy arrays as well, as soon as
        # https://github.com/cupy/cupy/issues/2616 is resolved
        self._writeable = False

    @property
    def readonly(self):
        """Indicates whether the AnyArray instance is read-only.

        Note
        ----
        For `cupy.ndarray`, write protection is not fully supported yet.

        Returns
        -------
        bool
            True if the array is read-only, False otherwise.

        Examples
        --------
        >>> arr = AnyArray(np.array([1, 2, 3]))
        >>> arr.readonly
        False
        >>> arr.lock()
        >>> arr.readonly
        True

        """
        return not self._writeable

    def at(self, device_id, *, check_fail=True):
        """Returns a copy of the AnyArray on the specified device.

        This method ensures that the AnyArray instance is located on the
        requested device (CPU or GPU). If the array is already on the target
        device, it is returned as-is. Otherwise, a copy is made and transferred
        to the target device.

        If the source array is read-only, the copied array on the new device
        will also be locked as read-only.

        Parameters
        ----------
        device_id : int
            The target device ID:
            - `-1` for CPU (`numpy.ndarray`)
            - `>= 0` for a specific GPU (`cupy.ndarray`)
        check_fail : bool, optional
            If True, additional debugging checks are performed to catch
            unexpected device transfers (default: True).

        Returns
        -------
        AnyArray
            A new AnyArray instance located on the specified device.

        Raises
        ------
        RuntimeError
            If `_config["fail_on_device_copy"]` is enabled and an unexpected device
            transfer occurs.

        Examples
        --------
        >>> arr = AnyArray(np.array([1, 2, 3]))
        >>> arr.device_id
        -1  # Currently on CPU

        >>> gpu_arr = arr.at(0)  # Move to GPU (if available)
        >>> gpu_arr.device_id
        0  # Now on GPU

        >>> cpu_arr = gpu_arr.at(-1)  # Move back to CPU
        >>> cpu_arr.device_id
        -1  # Now on CPU again

        """
        myassert(isinstance(device_id, int))
        origin, target = self._device_id, device_id
        if origin != target:
            s = f"AnyArray copy {self._device_id} -> {device_id}: shape {self._shape}"
            if self.shape != tuple():
                logger.debug(s)
            if _config["break_on_device_copy"] and check_fail:
                breakpoint()
            if _config["fail_on_device_copy"] and check_fail and self.shape != tuple():
                raise RuntimeError("Unexpected Device Copy:" + s)
        if origin == target:  # no transfer
            return self
        elif target == -1:  # gpu->cpu
            res = AnyArray(cupy.asnumpy(self._val))
        elif target > -1:  # cpu/gpu->gpu
            assert_device_available()
            with cupy.cuda.Device(device_id):
                arr = cupy.asarray(self._val)
            myassert(arr.device.id == device_id)
            res = AnyArray(arr)
        else:
            raise ValueError(f"device_id needs to be >= -1, got {device_id}")
        if self.readonly:
            res.lock()
        return res

    def asnumpy(self):
        """Returns a NumPy representation of the AnyArray.

        This method ensures that the array is located on the CPU
        (`numpy.ndarray`). If the `AnyArray` is already on the CPU, it is
        returned as-is. Otherwise, it is transferred from GPU to CPU.

        If the AnyArray is read-only, the returned NumPy array will also be
        flagged as read-only.

        Returns
        -------
        numpy.ndarray
            The content of the AnyArray as NumPy array.

        """
        res = self.at(-1).val
        if self.readonly:
            res.flags.writeable = False
        return res

    def view(self, *args, **kwargs):
        """Returns a new view of the array with the specified shape or type.

        This method creates a view of the underlying data without making a copy.
        It behaves similarly to `numpy.ndarray.view()`, returning an AnyArray
        instance wrapping the new view.

        Views share memory with the original array; modifying one modifies the
        other.

        Parameters
        ----------
        See numpy or cupy documentation.

        Returns
        -------
        AnyArray
            A new AnyArray instance representing the view.

        Examples
        --------
        >>> arr = AnyArray(np.array([[1, 2], [3, 4]]))
        >>> view = arr.view()
        >>> view is arr
        False  # Different object, but shares memory

        """
        return AnyArray(self._val.view(*args, **kwargs))

    def copy(self):
        """Returns a deep copy of the AnyArray.

        This method creates an independent copy of the array, ensuring that
        modifications to the copy do not affect the original.

        The copy is always writeable, even if the original AnyArray is
        read-only.

        Returns
        -------
        AnyArray
            A new AnyArray instance containing a copy of the original data.

        Examples
        --------
        >>> arr = AnyArray(np.array([1, 2, 3]))
        >>> arr_copy = arr.copy()
        >>> (arr_copy.val is arr.val)
        False  # The data is copied, not shared

        """
        return AnyArray(self._val.copy())

    # ---Properties, etc.---
    @property
    def val(self):
        """Returns the underlying array.

        This property provides access to the raw `numpy.ndarray` or
        `cupy.ndarray` wrapped by `AnyArray`.

        Returns
        -------
        numpy.ndarray or cupy.ndarray
            The wrapped array.

        """
        return self._val

    @property
    def device_id(self):
        """Returns the device ID where the array is stored.

        - `-1` represents a CPU (NumPy) array.
        - A non-negative integer represents a GPU (CuPy) array on the
          corresponding GPU device.

        Returns
        -------
        int
            The device ID of the array.

        """
        return self._device_id

    @property
    def shape(self):
        return self._val.shape

    @property
    def ndim(self):
        return len(self._val.shape)

    @property
    def size(self):
        if self._val.shape == tuple():
            return 1
        return reduce(mul, self._val.shape)

    @property
    def nbytes(self):
        return self._val.nbytes

    @property
    def strides(self):
        return self._val.strides

    @property
    def dtype(self):
        return self._val.dtype

    def astype(self, *args, **kwargs):
        return AnyArray(self._val.astype(*args, **kwargs))

    # ---Helpers---
    def _preprocess_index(self, index):
        if isinstance(index, tuple):
            out = []
            for ind in index:
                if isinstance(index, ALLOWED_WRAPPEES):
                    raise TypeError("Need to wrap numpy and cupy arrays into AnyArrays")
                if isinstance(ind, AnyArray):
                    if self.device_id != ind.device_id:
                        raise RuntimeError("device id mismatch")
                    ind = ind._val
                out.append(ind)
            index = tuple(out)
        elif isinstance(index, AnyArray):
            if self.device_id != index.device_id:
                raise RuntimeError("device id mismatch")
            index = index._val
        elif isinstance(index, ALLOWED_WRAPPEES):
            raise TypeError("Need to wrap numpy and cupy arrays into AnyArrays")
        return index

    def __getitem__(self, index):
        index = self._preprocess_index(index)
        return AnyArray(self._val[index])

    def __setitem__(self, index, value):
        if self.readonly:
            raise ValueError("assignment destination is read-only")
        if isinstance(value, AnyArray):
            value = value._val
        index = self._preprocess_index(index)
        if not isinstance(value, (int, float, complex)) and type(value) is not type(self._val):
            raise TypeError(f"Expected: {type(self._val)}, got: {type(value)}")
        self._val[index] = value

    # ---Point-wise nonlinearities---
    @staticmethod
    def _prep_args(args, kwargs, device_id=-1):
        for arg in args + tuple(kwargs.values()):
            if not (arg is None or np.isscalar(arg) or arg.jac is None):
                raise TypeError("bad argument")
        argstmp = tuple(arg if arg is None or np.isscalar(arg)
                        else arg._val.at(device_id)._val for arg in args)
        kwargstmp = {key: val if val is None or np.isscalar(val)
                     else val._val.at(device_id)._val
                     for key, val in kwargs.items()}
        return argstmp, kwargstmp

    def ptw(self, op, *args, **kwargs):
        """Performs an element-wise operation on the array.

        This method applies a specified pointwise operation to the array, using
        the corresponding function from :mod:`nifty.cl.pointwise`.

        Parameters
        ----------
        op : str
            The name of the pointwise operation (e.g., `exp`).
        *args : tuple
            Additional positional arguments for the operation.
        **kwargs : dict
            Additional keyword arguments for the operation.

        Returns
        -------
        AnyArray
            A new `AnyArray` instance with the result of the operation.

        """
        from .pointwise import ptw_dict
        argstmp, kwargstmp = self._prep_args(args, kwargs, self._device_id)
        return AnyArray(ptw_dict[op][0](self._val, *argstmp, **kwargstmp))

    def ptw_with_deriv(self, op, *args, **kwargs):
        """Performs element-wise operation and derivative simultaneously.

        This method applies a specified pointwise operation to the array and
        also computes its derivative, using the corresponding function from
        :mod:`nifty.cl.pointwise`.

        Parameters
        ----------
        op : str
            The name of the pointwise operation (e.g., `exp`).
        *args : tuple
            Additional positional arguments for the operation.
        **kwargs : dict
            Additional keyword arguments for the operation.

        Returns
        -------
        tuple of AnyArray
            A tuple containing:
            - The result of the operation (`AnyArray`).
            - The derivative of the operation (`AnyArray`).

        Examples
        --------
        >>> arr = AnyArray(np.array([4, 9, 16]))
        >>> res, deriv = arr.ptw_with_deriv("sqrt")
        >>> res.val
        array([2., 3., 4.])
        >>> deriv.val
        array([0.25, 0.16666667, 0.125])

        """
        from .pointwise import ptw_dict
        argstmp, kwargstmp = self._prep_args(args, kwargs, self._device_id)
        return tuple(map(AnyArray, ptw_dict[op][1](self._val, *argstmp, **kwargstmp)))

    # ---Numeric properties---
    @property
    def real(self):
        return AnyArray(self._val.real)

    @property
    def imag(self):
        return AnyArray(self._val.imag)

    # ---Operations---
    def norm(self, ord=2):
        res = np.linalg.norm(self._val.reshape(-1), ord=ord)
        return _make_scalar_if_scalar(res)

    def vdot(self, x):
        x = x.at(self.device_id)
        if self._device_id == -1:
            return cpu_vdot(self._val, x._val)
        return _make_scalar_if_scalar(cupy.vdot(self._val, x._val))

    @property
    def T(self):
        return np.transpose(self)

    @property
    def H(self):
        return np.transpose(self.conj())

    # ---Numpy ufunc and array_function interface---
    def _unify_device_ids_and_get_val(self, args, kwargs):
        # TODO: Leverage Cupy's support for peer-to-peer memory access.
        device_id = None

        args2 = []
        for vv in args:
            if isinstance(vv, AnyArray):
                if device_id is None:
                    device_id = vv._device_id
                vv = vv.at(device_id)._val
            if isinstance(vv, (tuple, list)):
                tmp = []
                for ww in vv:
                    if isinstance(ww, AnyArray):
                        if device_id is None:
                            device_id = ww._device_id
                        ww = ww.at(device_id)._val
                    tmp.append(ww)
                vv = tuple(tmp)
            args2.append(vv)

        kwargs2 = {}
        for ii, vv in kwargs.items():
            if ii == "out":
                for vvv in vv:
                    if isinstance(vv, AnyArray) and device_id is not None \
                       and vvv.device_id != device_id:
                        raise RuntimeError("Specifying 'out' on a different "
                                           "device is not supported (yet).")
            if isinstance(vv, AnyArray):
                if device_id is None:
                    device_id = vv._device_id
                vv = vv.at(device_id)._val
            if isinstance(vv, (tuple, list)):
                tmp = []
                for ww in vv:
                    if isinstance(ww, AnyArray):
                        if device_id is None:
                            device_id = ww._device_id
                        ww = ww.at(device_id)._val
                    tmp.append(ww)
                vv = tuple(tmp)
            kwargs2[ii] = vv

        return args2, kwargs2

    @staticmethod
    def _check_responsibility(lst):
        any_array_present = False
        for x in lst:
            if isinstance(x, ALLOWED_WRAPPEES):
                return False
            elif isinstance(x, (tuple, list)):
                for xx in x:
                    if isinstance(xx, ALLOWED_WRAPPEES):
                        return False
                    if isinstance(xx, AnyArray):
                        any_array_present = True
            elif isinstance(x, AnyArray):
                any_array_present = True
        return any_array_present

    @staticmethod
    def _wrap_result(obj, out=None):
        if obj is NotImplemented:
            warn("Falling back to inefficient way to determine responsibility "
                 "for AnyArray operations. Please report.")
            return NotImplemented
        # TODO: Add make_scalar_if_scalar here
        if out is not None:
            return out
        if isinstance(obj, ALLOWED_WRAPPEES):
            return AnyArray(obj)
        if isinstance(obj, tuple):
            return tuple(AnyArray(x) if isinstance(x, ALLOWED_WRAPPEES)
                         else x for x in obj)
        return obj

    def __array_ufunc__(self, ufunc, method, *args, **kwargs):
        if not self._check_responsibility(chain(args, kwargs.values())):
            return NotImplemented
        args2, kwargs2 = self._unify_device_ids_and_get_val(args, kwargs)
        return self._wrap_result(getattr(ufunc, method)(*args2, **kwargs2),
                                 out=kwargs.get("out", None))

    def __array_function__(self, func, types, args, kwargs):
        if any(not issubclass(t, AnyArray) for t in types):
            return NotImplemented
        if not self._check_responsibility(chain(args, kwargs.values())):
            return NotImplemented
        args2, kwargs2 = self._unify_device_ids_and_get_val(args, kwargs)
        return self._wrap_result(func(*args2, **kwargs2),
                                 out=kwargs.get("out", None))

    # ---Printing functions---
    def __repr__(self):
        return f"<nifty.cl.AnyArray: device_id={self._device_id}>"

    def __str__(self):
        return "\n".join(["nifty.cl.AnyArray instance",
                          f"- device_id = {self.device_id}",
                          f"- shape     = {self.shape}",
                          f"- strides   = {self.strides}",
                          f"- dtype     = {self.dtype}",
                          f"- size      = {self.size}",
                          f"- memory    = {self.nbytes*1e-6:.3f} MB",
                          ])

    def __bool__(self):
        return self._val.__bool__()


#---Operations with potential scalar output---
for op in ["sum", "prod", "mean", "var", "std", "all", "any", "min", "max",
           "conjugate", "reshape", "conj", "flatten"]:
    def func(op):
        def func2(self, *args, **kwargs):
            res = getattr(self._val, op)(*args, **kwargs)
            # Convention: simple types (float, int, complex, ...) are always on the host
            res = _make_scalar_if_scalar(res)
            return res if np.isscalar(res) else AnyArray(res)
        return func2
    setattr(AnyArray, op, func(op))


# ---Inplace operations---
for op in ["__iadd__", "__isub__", "__imul__", "__idiv__",
           "__itruediv__", "__ifloordiv__", "__ipow__"]:
    def func(op):
        def func2(self, other):
            if not isinstance(other, AnyArray):
                return NotImplemented
            if self.device_id != other.device_id:
                raise RuntimeError(f"Cannot compute {op} if AnyArrays are stored on different devices. "
                                   f"Got: {self.device_id} and {other.device_id}.")
            if self._writeable:
                return AnyArray(getattr(self._val, op)(other._val))
            raise TypeError("AnyArray is readonly")
        return func2
    setattr(AnyArray, op, func(op))



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2021-2022 Max-Planck-Society
# Copyright(C) 2025 Philipp Arras
#
# Author: Philipp Arras

from functools import partial
from types import SimpleNamespace
from warnings import warn

import numpy as np

from ..any_array import AnyArray, device_available
from .energy_operators import LikelihoodEnergyOperator
from .linear_operator import LinearOperator
from .operator import Operator

__all__ = ["JaxOperator", "JaxLikelihoodEnergyOperator", "JaxLinearOperator"]


def _jax2anyarray(obj):
    if isinstance(obj, dict):
        return {kk: _jax2anyarray(vv) for kk, vv in obj.items()}
    elif device_available():
        import cupy
        cu = cupy.from_dlpack(obj)
        return AnyArray(cu)
    else:
        return AnyArray(np.array(obj))


def _anyarray2jax(obj):
    import jax
    if isinstance(obj, AnyArray):
        if obj.device_id == -1:
            return jax.numpy.array(obj._val)
        else:
            return jax.dlpack.from_dlpack(obj._val)
    elif isinstance(obj, dict):
        return {kk: _anyarray2jax(vv) for kk, vv in obj.items()}


class JaxOperator(Operator):
    """Wrap a jax function as nifty operator.

    Parameters
    ----------
    domain : DomainTuple or MultiDomain
        Domain of the operator.

    target : DomainTuple or MultiDomain
        Target of the operator.

    func : callable
        The jax function that is evaluated by the operator. It has to be
        implemented in terms of `jax.numpy` calls. If `domain` is a
        `MultiDomain`, `func` takes a `dict` as argument and like-wise for the
        target.

    Note
    ----
    Contrary to the convention in the rest of nifty, this operator returns
    Fields on device if a device is available. Normally, nifty operators return
    on the same device_id as their input has been.
    """
    def __init__(self, domain, target, func):
        import jax

        from ..sugar import makeDomain
        self._domain = makeDomain(domain)
        self._target = makeDomain(target)
        self._func = jax.jit(func)
        self._bwd = jax.jit(lambda x, y: jax.vjp(func, x)[1](y)[0])
        self._fwd = jax.jit(lambda x, y: jax.jvp(self._func, (x,), (y,))[1])

    def apply(self, x):
        from ..multi_domain import MultiDomain
        from ..sugar import is_linearization, makeField
        self._check_input(x)
        if is_linearization(x):
            # TODO: Adapt the Linearization class to handle value_and_grad
            # calls. Computing the pass through the function thrice (once now
            # and twice when differentiating) is redundant and inefficient.
            res = self._func(x.val.asnumpy())
            bwd = partial(self._bwd, _anyarray2jax(x.val.val))
            fwd = partial(self._fwd, _anyarray2jax(x.val.val))
            jac = JaxLinearOperator(self._domain, self._target, fwd, func_T=bwd)
            return x.new(makeField(self._target, _jax2anyarray(res)), jac)
        res = _jax2anyarray(self._func(_anyarray2jax(x.val)))
        if isinstance(res, dict):
            if not isinstance(self._target, MultiDomain):
                raise TypeError(("Jax function returns a dictionary although the "
                                 "target of the operator is a DomainTuple."))
            if set(res.keys()) != set(self._target.keys()):
                raise ValueError(("Keys do not match:\n"
                                  f"Target keys: {self._target.keys()}\n"
                                  f"Jax function returns: {res.keys()}"))
            for kk in res.keys():
                self._check_shape(self._target[kk].shape, res[kk].shape)
        else:
            if isinstance(self._target, MultiDomain):
                raise TypeError(("Jax function does not return a dictionary "
                                 "although the target of the operator is a "
                                 "MultiDomain."))
            self._check_shape(self._target.shape, res.shape)
        return makeField(self._target, res)

    @staticmethod
    def _check_shape(shp_tgt, shp_jax):
        if shp_tgt != shp_jax:
            raise ValueError(("Output shapes do not match:\n"
                              f"Target shape is\t\t{shp_tgt}\n"
                              f"Jax function returns\t{shp_jax}"))

    def _simplify_for_constant_input_nontrivial(self, c_inp):
        func2 = lambda x: self._func({**x, **c_inp.asnumpy()})
        dom = {kk: vv for kk, vv in self._domain.items() if kk not in c_inp.keys()}
        return None, JaxOperator(dom, self._target, func2)


class JaxLinearOperator(LinearOperator):
    """Wrap a jax function as nifty linear operator.

    Parameters
    ----------
    domain : DomainTuple or MultiDomain
        Domain of the operator.

    target : DomainTuple or MultiDomain
        Target of the operator.

    func : callable
        The jax function that is evaluated by the operator. It has to be
        implemented in terms of `jax.numpy` calls. If `domain` is a
        `MultiDomain`, `func` takes a `dict` as argument and like-wise for the
        target.

    func_T : callable
        The jax function that implements the transposed action of the operator.
        If None, jax computes the adjoint. Note that this is *not* the adjoint
        action. Default: None.

    domain_dtype:
        Needs to be set if `func_transposed` is None. Otherwise it does not have
        an effect. Dtype of the domain. If `domain` is a `MultiDomain`,
        `domain_dtype` is supposed to be a dictionary. Default: None.

    Note
    ----
    It is the user's responsibility that func is actually a linear function. The
    user can double check this with the help of
    `nifty.cl.extra.check_linear_operator`.
    """
    def __init__(self, domain, target, func, domain_dtype=None, func_T=None):
        import jax

        from ..domain_tuple import DomainTuple
        from ..sugar import makeDomain
        domain = makeDomain(domain)
        if domain_dtype is not None and func_T is None:
            if isinstance(domain, DomainTuple):
                inp = SimpleNamespace(shape=domain.shape, dtype=domain_dtype)
            else:
                inp = {kk: SimpleNamespace(shape=domain[kk].shape, dtype=domain_dtype[kk])
                       for kk in domain.keys()}
            func_T = jax.jit(lambda x: jax.linear_transpose(func, inp)(x)[0])
        elif domain_dtype is None and func_T is not None:
            pass
        else:
            raise ValueError("Either domain_dtype or func_T have to be not None.")
        self._domain = makeDomain(domain)
        self._target = makeDomain(target)
        self._func = func
        self._func_T = func_T
        self._capability = self.TIMES | self.ADJOINT_TIMES

    def apply(self, x, mode):
        from ..sugar import makeField
        self._check_input(x, mode)
        if mode == self.TIMES:
            fx = self._func(_anyarray2jax(x.val))
            return makeField(self._target, _jax2anyarray(fx))
        # TODO: make conjugate part of the jax operation
        fx = self._func_T(_anyarray2jax(x.conjugate().val))
        return makeField(self._domain, _jax2anyarray(fx)).conjugate()


class JaxLikelihoodEnergyOperator(LikelihoodEnergyOperator):
    """Wrap a jax function as nifty likelihood energy operator.

    Parameters
    ----------
    domain : DomainTuple or MultiDomain
        Domain of the operator.

    func : callable
        The jax function that is evaluated by the operator. It has to be
        implemented in terms of `jax.numpy` calls. If `domain` is a
        `MultiDomain`, `func` takes a `dict` as argument and like-wise for the
        target. It needs to map to a scalar.

    transformation : Operator, optional
        Coordinate transformation to Euclidean space.

    sampling_dtype : dtype, optional
        The dtype that shall be used for drawing samples from the metric of the
        likelihood.
    """
    def __init__(self, domain, func, transformation=None, sampling_dtype=None):
        import jax

        from ..sugar import makeDomain
        self._domain = makeDomain(domain)
        self._func = jax.jit(func)
        self._val_and_grad = jax.jit(jax.value_and_grad(func))
        self._dt = sampling_dtype
        self._trafo = transformation
        warn("JaxLikelihoodEnergyOperator does not support normalized residuals "
             "yet. This mean that the inference works but the data residuals do "
             "not show up in e.g. `ift.minisanity`.")
        super(JaxLikelihoodEnergyOperator, self).__init__(None, None)

    def get_transformation(self):
        if self._trafo is None:
            s = self.__name__ + " was instantiated without `transformation`"
            raise RuntimeError(s)
        return self._dt, self._trafo

    def apply(self, x):
        from ..sugar import is_linearization, makeField
        from .simple_linear_operators import VdotOperator

        self._check_input(x)
        lin = is_linearization(x)
        val = x.val.val if lin else x.val
        if not lin:
            return makeField(self._target, _jax2anyarray(self._func(_anyarray2jax(val))))
        res, grad = self._val_and_grad(_anyarray2jax(val))
        jac = VdotOperator(makeField(self._domain, _jax2anyarray(grad)))
        res = x.new(makeField(self._target, _jax2anyarray(res)), jac)
        if not x.want_metric:
            return res
        # TODO: Add jax version of metric
        return res.add_metric(self.get_metric_at(x.val))

    def _simplify_for_constant_input_nontrivial(self, c_inp):
        func2 = lambda x: self._func({**x, **_anyarray2jax(c_inp.val)})
        dom = {kk: vv for kk, vv in self._domain.items() if kk not in c_inp.keys()}
        _, trafo = self._trafo.simplify_for_constant_input(c_inp)
        if isinstance(self._dt, dict):
            dt = {kk: self._dt[kk] for kk in dom.keys()}
        else:
            dt = self._dt
        return None, JaxLikelihoodEnergyOperator(dom, func2, trafo, dt)
